# =============================================================================
# Minet FileSystem Utilities
# =============================================================================
#
# Multiple helper functions related to reading and writing files.
#
import gzip
from os import makedirs
from os.path import basename, join, splitext, abspath, normpath, dirname
from ural import get_hostname, get_normalized_hostname
from quenouille import NamedLocks

from minet.exceptions import FilenameFormattingError
from minet.utils import md5, PseudoFStringFormatter


def read_potentially_gzipped_path(path, encoding='utf-8'):
    open_fn = open
    flag = 'r'

    if path.endswith('.gz'):
        open_fn = gzip.open
        flag = 'rt'

    with open_fn(path, flag, encoding=encoding, errors='replace') as f:
        return f.read()


class FolderStrategy(object):
    def __call__(self):
        raise NotImplementedError

    @staticmethod
    def from_name(name):
        if name == 'flat':
            return FlatFolderStrategy()

        if name == 'hostname':
            return HostnameFolderStrategy()

        if name == 'normalized-hostname':
            return NormalizedHostnameFolderStrategy()

        if name.startswith('prefix-'):
            length = name.split('prefix-')[-1]

            try:
                length = int(length)
            except ValueError:
                raise TypeError

            if length <= 0:
                raise TypeError

            return PrefixFolderStrategy(length)

        raise TypeError


class FlatFolderStrategy(FolderStrategy):
    def __call__(self, filename, **kwargs):
        return filename


class PrefixFolderStrategy(FolderStrategy):
    def __init__(self, length):
        self.length = length

    def __call__(self, filename, **kwargs):
        base = basename(filename).split('.', 1)[0]

        return join(base[:self.length], filename)


class HostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_hostname(url)

        if not hostname:
            hostname = 'unknown-host'

        return join(hostname, filename)


class NormalizedHostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_normalized_hostname(
            url,
            normalize_amp=False,
            strip_lang_subdomains=True,
            infer_redirection=False
        )

        if not hostname:
            hostname = 'unknown-host'

        return join(hostname, filename)


class FilenameBuilder(object):
    def __init__(self, folder_strategy=None, template=None):
        self.folder_strategy = None

        if folder_strategy is not None:
            self.folder_strategy = FolderStrategy.from_name(folder_strategy)

        self.formatter = PseudoFStringFormatter()
        self.template = template

    def __call__(self, url=None, filename=None, ext=None, formatter_kwargs={},
                 compressed=False):
        original_ext = None

        if filename is None:
            base = md5(url)
        else:
            base, original_ext = splitext(filename)

        # We favor the extension found in given filename, else we fallback
        # on the provided one if any (usually inferred from http response)
        ext = original_ext if original_ext else (ext or '')

        if self.template is not None:
            try:
                filename = self.formatter.format(
                    self.template,
                    value=base,
                    ext=ext,
                    **formatter_kwargs
                )
            except Exception as e:
                raise FilenameFormattingError(reason=e, template=self.template)
        else:
            filename = base + ext

        if self.folder_strategy:
            filename = self.folder_strategy(filename, url=url)

        if compressed:
            filename += '.gz'

        return filename


class ThreadSafeFilesWriter(object):
    def __init__(self, root_directory=''):
        self.root_directory = root_directory
        self.folder_locks = NamedLocks()
        self.file_locks = NamedLocks()

    def resolve(self, filename, relative=False):
        full_path = join(self.root_directory, filename)

        if relative:
            return normpath(full_path)

        return abspath(full_path)

    def makedirs(self, directory):
        if not directory:
            return

        # TODO: cache
        with self.folder_locks[directory]:
            makedirs(directory, exist_ok=True)

    def write(self, filename, contents, binary=True, compress=False):
        if binary and not isinstance(contents, bytes):
            raise TypeError('contents must be bytes if binary=True')

        if not binary and not isinstance(contents, str):
            raise TypeError('contents must be str if binary=False')

        if compress and not binary:
            raise NotImplementedError

        filename = self.resolve(filename)
        directory = dirname(filename)

        # NOTE: Could have prefix-free locking as a bonus...
        self.makedirs(directory)

        open_kwargs = {
            'mode': 'wb' if binary else 'w'
        }

        if not binary:
            open_kwargs['encoding'] = 'utf-8'

        if compress:
            open_fn = gzip.open
        else:
            open_fn = open

        with self.file_locks[filename]:
            with open_fn(filename, **open_kwargs) as f:
                f.write(contents)



# =============================================================================
# Minet Apply Scraper Function
# =============================================================================
#
# Function taking a scraper definition and applying its logic recursively
# to yield its result.
#
import textwrap
import soupsieve
from bs4 import Tag
from ebbe import getpath
from functools import partial

from minet.scrape.std import get_default_evaluation_context, get_display_text
from minet.scrape.constants import EXTRACTOR_NAMES
from minet.scrape.utils import get_sel, get_iterator
from minet.scrape.exceptions import (
    ScraperEvalError,
    ScraperEvalTypeError,
    ScraperEvalNoneError,
    NotATableError
)

DEFAULT_CONTEXT = {}
DATA_TYPES = (str, int, float, bool, list, dict)

nested_getter = partial(getpath, split_char='.', parse_indices=True, attributes=True)


def is_list_of_tags(value):
    if not isinstance(value, list):
        return False

    return all(isinstance(tag, Tag) for tag in value)


def is_valid_iterator_eval_output(value):
    return isinstance(value, str) or is_list_of_tags(value)


def merge_contexts(global_context, local_context):
    if global_context is None:
        return local_context

    for k, v in global_context.items():
        if k not in local_context:
            local_context[k] = v

    return local_context


def extract(element, extractor_name):
    if extractor_name == 'text':
        return element.get_text().strip()

    if extractor_name == 'display_text':
        return get_display_text(element)

    if extractor_name == 'html' or extractor_name == 'inner_html':
        return element.decode_contents().strip()

    if extractor_name == 'outer_html':
        return str(element).strip()

    raise TypeError('Unknown "%s" extractor' % extractor_name)


class EvaluationScope(object):
    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, key):
        return self.__dict__.get(key)


EVAL_CONTEXT = get_default_evaluation_context()


# NOTE: this is not threadsafe, but it does not have to be
def eval_expression(expression, element=None, elements=None, value=None,
                    context=None, root=None, scope=None,
                    path=None, expect=None, check=None, allow_none=False):

    if callable(expression):
        try:
            result = expression(
                element=element,
                elements=elements,
                value=value,
                context=context,
                root=root
            )
        except Exception as e:
            raise ScraperEvalError(
                reason=e,
                path=path,
                expression=expression
            )
    else:

        # Local variables
        EVAL_CONTEXT['element'] = element
        EVAL_CONTEXT['elements'] = elements
        EVAL_CONTEXT['value'] = value

        # Context
        EVAL_CONTEXT['context'] = context
        EVAL_CONTEXT['root'] = root
        EVAL_CONTEXT['scope'] = scope

        if '\n' in expression:

            # Multiline expression
            scope = {}

            wrapped_expression = 'def __run__():\n%s\n__return_value__ = __run__()' % textwrap.indent(expression, '  ')

            try:
                exec(wrapped_expression, EVAL_CONTEXT, scope)
                result = scope['__return_value__']
            except Exception as e:
                raise ScraperEvalError(
                    reason=e,
                    path=path,
                    expression=expression
                )
        else:

            # Simple expression
            try:
                result = eval(expression, EVAL_CONTEXT, None)
            except Exception as e:
                raise ScraperEvalError(
                    reason=e,
                    path=path,
                    expression=expression
                )

    if not allow_none and result is None:
        raise ScraperEvalNoneError(
            path=path,
            expression=expression
        )

    if (expect is not None or check is not None) and result is not None:
        if not (check(result) if check else isinstance(result, expect)):
            raise ScraperEvalTypeError(
                path=path,
                expression=expression,
                expected=expect,
                got=result
            )

    return result


def tabulate(element, headers_inference='th', headers=None, path=None):
    if element.name != 'table':
        raise NotATableError(path=path)

    def generator():
        nonlocal headers

        body = element.find('tbody', recursive=False)
        head = element.find('thead', recursive=False)

        if body is None:
            body = element

        if head is None:
            head = element

        trs = body.select('tr:has(td)', recursive=False)
        ths = head.select('tr > th', recursive=False)

        if headers is None:
            if headers_inference == 'th':
                headers = [th.get_text() for th in ths]

        if headers is not None:
            for tr in trs:
                yield {headers[i]: td.get_text() for i, td in enumerate(tr.find_all('td', recursive=False))}
        else:
            for tr in trs:
                yield [td.get_text() for td in tr.find_all('td', recursive=False)]

    return generator()


def interpret_scraper(scraper, element, root=None, context=None, path=[], scope=None):
    if scope is None:
        scope = EvaluationScope()

    # Is this a tail call of item?
    if isinstance(scraper, str):
        if scraper in EXTRACTOR_NAMES:
            return extract(element, scraper)

        return element.get(scraper)

    sel = get_sel(scraper)
    iterator = get_iterator(scraper)

    # First we need to solve local selection
    if sel is not None:
        element = soupsieve.select_one(sel, element)
    elif 'sel_eval' in scraper:

        evaluated_sel = eval_expression(
            scraper['sel_eval'],
            element=element,
            elements=[],
            context=context,
            root=root,
            path=path + ['sel_eval'],
            expect=(Tag, str),
            allow_none=True,
            scope=scope
        )

        if isinstance(evaluated_sel, str):
            element = soupsieve.select_one(evaluated_sel, element)
        else:
            element = evaluated_sel

    if element is None:
        return None

    # Then we need to solve iterator
    single_value = True

    if iterator is not None:
        single_value = False
        elements = soupsieve.select(iterator, element)
    elif 'iterator_eval' in scraper:
        single_value = False
        evaluated_elements = eval_expression(
            scraper['iterator_eval'],
            element=element,
            elements=[],
            context=context,
            root=root,
            path=path + ['iterator_eval'],
            check=is_valid_iterator_eval_output,
            scope=scope
        )

        if isinstance(evaluated_elements, str):
            elements = soupsieve.select(evaluated_elements, element)
        else:
            elements = evaluated_elements
    else:
        elements = [element]

    # Handling local context
    if 'set_context' in scraper:
        local_context = {}

        for k, field_scraper in scraper['set_context'].items():
            local_context[k] = interpret_scraper(
                field_scraper,
                element,
                root=root,
                context=context,
                path=path + ['set_context', k],
                scope=scope
            )

        context = merge_contexts(context, local_context)

    # Actual iteration
    acc = None if single_value else []

    already_seen = set() if 'uniq' in scraper and not single_value else None

    for element in elements:
        value = None

        # Do we have fields?
        if 'fields' in scraper:
            value = {}

            for k, field_scraper in scraper['fields'].items():
                value[k] = interpret_scraper(
                    field_scraper,
                    element,
                    root=root,
                    context=context,
                    path=path + ['fields', k],
                    scope=scope
                )

        # Do we have a scalar?
        elif 'item' in scraper:

            # Default value is text
            value = interpret_scraper(
                scraper['item'],
                element,
                root=root,
                context=context,
                path=path + ['item'],
                scope=scope
            )

        else:

            if 'attr' in scraper:
                value = element.get(scraper['attr'])
            elif 'extract' in scraper:
                value = extract(element, scraper['extract'])
            elif 'get_context' in scraper:
                value = nested_getter(context, scraper['get_context'])
            elif 'default' not in scraper:

                # Default value is text
                value = extract(element, 'text')

            # Eval?
            if 'eval' in scraper:
                value = eval_expression(
                    scraper['eval'],
                    element=element,
                    elements=elements,
                    value=value,
                    context=context,
                    root=root,
                    path=path + ['eval'],
                    expect=DATA_TYPES,
                    allow_none=True,
                    scope=scope
                )

        # Default value after all?
        if 'default' in scraper and value is None:
            value = scraper['default']

        if single_value:
            acc = value
        else:

            # Filtering?
            if 'filter_eval' in scraper:
                passed_filter = eval_expression(
                    scraper['filter_eval'],
                    element=element,
                    elements=elements,
                    value=value,
                    context=context,
                    root=root,
                    path=path + ['filter_eval'],
                    expect=bool,
                    allow_none=True,
                    scope=scope
                )

                if not passed_filter:
                    continue

            if 'filter' in scraper:
                filtering_clause = scraper['filter']

                if filtering_clause is True and not value:
                    continue

                if isinstance(filtering_clause, str) and not nested_getter(value, filtering_clause):
                    continue

            if 'uniq' in scraper:
                uniq_clause = scraper['uniq']
                k = value

                if uniq_clause is True and value in already_seen:
                    continue

                if isinstance(uniq_clause, str):
                    k = nested_getter(value, uniq_clause)

                    if k in already_seen:
                        continue

                already_seen.add(k)

            acc.append(value)

    return acc



# =============================================================================
# Minet Twitter Users CLI Action
# =============================================================================
#
# Logic of the `tw users` action.
#
import casanova
from twitwi import (
    TwitterWrapper,
    normalize_user,
    format_user_as_csv_row
)
from twitter import TwitterHTTPError
from twitwi.constants import USER_FIELDS
from ebbe import as_chunks

from minet.cli.utils import LoadingBar


def twitter_users_action(cli_args):

    wrapper = TwitterWrapper(
        cli_args.access_token,
        cli_args.access_token_secret,
        cli_args.api_key,
        cli_args.api_secret_key
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=USER_FIELDS
    )

    loading_bar = LoadingBar(
        desc='Retrieving users',
        total=cli_args.total,
        unit='user'
    )

    for chunk in as_chunks(100, enricher.cells(cli_args.column, with_rows=True)):
        users = ','.join(row[1] for row in chunk)

        if cli_args.ids:
            wrapper_args = {'user_id': users}
            key = 'id'
        else:
            wrapper_args = {'screen_name': users}
            key = 'screen_name'

        try:
            result = wrapper.call(['users', 'lookup'], **wrapper_args)
        except TwitterHTTPError as e:
            if e.e.code == 404:
                for row, user in chunk:
                    enricher.writerow(row, user_row)
            else:
                raise e

        if result is not None:
            indexed_result = {}

            for user in result:
                user = normalize_user(user)
                user_row = format_user_as_csv_row(user)
                indexed_result[user[key]] = user_row

            for row, user in chunk:
                user_row = indexed_result.get(user)

                enricher.writerow(row, user_row)

        loading_bar.update(len(chunk))



# =============================================================================
# Minet Mediacloud Topic
# =============================================================================
#
# Function related to topics.
#
from minet.web import request_json
from minet.mediacloud.constants import (
    MEDIACLOUD_API_BASE_URL,
    MEDIACLOUD_DEFAULT_BATCH
)
from minet.mediacloud.utils import get_next_link_id
from minet.mediacloud.formatters import format_topic_story


def url_forge(token=None, topic_id=None, link_id=None, media_id=None,
              from_media_id=None):

    url = '%s/topics/%s/stories/list?key=%s&limit=%i' % (
        MEDIACLOUD_API_BASE_URL,
        topic_id,
        token,
        MEDIACLOUD_DEFAULT_BATCH
    )

    if link_id is not None:
        url += '&link_id=%s' % link_id

    if media_id is not None:
        url += '&media_id=%s' % media_id

    if from_media_id is not None:
        url += '&link_from_media_id=%s' % from_media_id

    return url


def mediacloud_topic_stories(pool, token, topic_id, link_id=None, media_id=None,
                             from_media_id=None, raw=False):

    while True:
        url = url_forge(
            token,
            topic_id=topic_id,
            link_id=link_id,
            media_id=media_id,
            from_media_id=from_media_id,
        )

        err, _, data = request_json(url, pool=pool)

        if err:
            raise err

        if 'stories' not in data or len(data['stories']) == 0:
            return

        next_link_id = get_next_link_id(data)

        for story in data['stories']:
            if not raw:
                story = format_topic_story(story, next_link_id)

            yield story

        if next_link_id is None:
            return

        link_id = next_link_id



# =============================================================================
# Minet Mediacloud Formatters
# =============================================================================
#
# Miscellaneous mediacloud-related formatter functions.
#
from casanova import namedrecord

from minet.mediacloud.utils import explode_tags
from minet.mediacloud.constants import (
    MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS,
    MEDIACLOUD_STORIES_CSV_HEADER,
    MEDIACLOUD_MEDIA_CSV_HEADER,
    MEDIACLOUD_FEED_CSV_HEADER
)

MediacloudTopicStory = namedrecord(
    'MediacloudTopicStory',
    MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS,
    boolean=['full_text_rss']
)

MediacloudStory = namedrecord(
    'MediacloudStory',
    MEDIACLOUD_STORIES_CSV_HEADER,
    plural=[
        'tags',
        'tag_sets',
        'tags_ids',
        'tag_sets_ids'
    ]
)

MediacloudMedia = namedrecord(
    'MediacloudMedia',
    MEDIACLOUD_MEDIA_CSV_HEADER,
    boolean=['is_healthy', 'is_monitored'],
    plural=[
        'tags',
        'tag_sets',
        'tags_ids',
        'tag_sets_ids'
    ]
)

MediacloudFeed = namedrecord(
    'MediacloudFeed',
    MEDIACLOUD_FEED_CSV_HEADER,
    boolean=['active']
)


def format_topic_story(story, next_link_id=None):
    return MediacloudTopicStory(
        story['guid'],
        story['stories_id'],
        story['title'],
        story['url'],
        story['language'],
        story['media_id'],
        story['media_name'],
        story['collect_date'],
        story['publish_date'],
        story['date_is_reliable'],
        story['facebook_share_count'],
        bool(story['full_text_rss']),
        story['inlink_count'],
        story['outlink_count'],
        story['media_inlink_count'],
        story['post_count'],
        story['snapshots_id'],
        story['timespans_id'],
        next_link_id
    )


def format_story(story, as_dict=False):
    tags = story.get('story_tags')

    return MediacloudStory(
        story['guid'],
        story['stories_id'],
        story['processed_stories_id'],
        story['title'],
        story['url'],
        story['language'],
        story['collect_date'],
        story['publish_date'],
        story['media_id'],
        story['media_name'],
        story['media_url'],
        *explode_tags(tags)
    )


def format_media(media, as_dict=False):
    tags = media.get('media_source_tags')

    return MediacloudMedia(
        media['media_id'],
        media['name'],
        media['url'],
        media['is_healthy'],
        media['is_monitored'],
        media.get('public_notes'),
        media['num_stories_90'],
        media['num_sentences_90'],
        media['start_date'],
        *explode_tags(tags)
    )


def format_feed(feed, as_dict=False):
    return MediacloudFeed(
        feed['name'],
        feed['url'],
        feed['feeds_id'],
        feed['type'],
        feed['media_id'],
        feed['active'],
        feed.get('last_attempted_download_time'),
        feed.get('last_new_story_time'),
        feed.get('last_successful_download_time')
    )



# =============================================================================
# Minet Scrape CLI Action
# =============================================================================
#
# Logic of the scrape action.
#
import csv
import sys
import ndjson
import casanova
from casanova import DictLikeRow
from termcolor import colored
from collections import namedtuple
from os.path import basename, isdir

from minet import Scraper
from minet.multiprocessing import LazyPool
from minet.exceptions import (
    DefinitionInvalidFormatError,
)
from minet.scrape.exceptions import (
    InvalidScraperError,
    CSSSelectorTooComplex,
    ScraperEvalError,
    ScraperEvalTypeError,
    ScraperEvalNoneError
)
from minet.cli.reporters import (
    report_scraper_validation_errors,
    report_scraper_evaluation_error
)
from minet.fs import read_potentially_gzipped_path
from minet.cli.utils import (
    die,
    dummy_csv_file_from_glob,
    create_report_iterator,
    LoadingBar
)
from minet.cli.constants import DEFAULT_CONTENT_FOLDER

ScrapeWorkerResult = namedtuple(
    'ScrapeWorkerResult',
    ['error', 'items']
)

PROCESS_SCRAPER = None


def init_process(definition, strain):
    global PROCESS_SCRAPER

    PROCESS_SCRAPER = Scraper(definition, strain=strain)


def worker(payload):
    row, headers, path, encoding, content, args = payload
    output_format, plural_separator = args

    # Reading from file
    if content is None:
        try:
            content = read_potentially_gzipped_path(path, encoding=encoding)
        except (FileNotFoundError, UnicodeDecodeError) as e:
            return ScrapeWorkerResult(e, None)

    # Building context
    context = {}

    if row:
        context['line'] = DictLikeRow(headers, row)

    if path:
        context['path'] = path
        context['basename'] = basename(path)

    # Attempting to scrape
    if output_format == 'csv':
        items = PROCESS_SCRAPER.as_csv_dict_rows(content, context=context, plural_separator=plural_separator)
    else:
        items = PROCESS_SCRAPER.as_records(content, context=context)

    # NOTE: errors will only be raised when we consume the generators created above
    try:
        items = list(items)
    except (ScraperEvalError, ScraperEvalTypeError, ScraperEvalNoneError) as error:
        return ScrapeWorkerResult(error, None)

    return ScrapeWorkerResult(None, items)


def scrape_action(cli_args):
    if cli_args.glob is None and cli_args.input_dir is None:
        cli_args.input_dir = DEFAULT_CONTENT_FOLDER

    # Parsing scraper definition
    try:
        scraper = Scraper(cli_args.scraper, strain=cli_args.strain)
    except DefinitionInvalidFormatError:
        die([
            'Unknown scraper format!',
            'It should be a JSON or YAML file.'
        ])
    except FileNotFoundError:
        die('Could not find scraper file!')
    except InvalidScraperError as error:
        print('Your scraper is invalid! You need to fix the following errors:', file=sys.stderr)
        print(file=sys.stderr)
        sys.stderr.write(report_scraper_validation_errors(error.validation_errors))
        die()
    except CSSSelectorTooComplex:
        die([
            'Your strainer\'s CSS selector %s is too complex.' % colored(cli_args.strain, 'blue'),
            'You cannot use relations to create a strainer.',
            'Try to simplify the selector you passed to --strain.'
        ])

    if cli_args.validate:
        print('Your scraper is valid.', file=sys.stderr)
        sys.exit(0)

    if scraper.headers is None and cli_args.format == 'csv':
        die([
            'Your scraper does not yield tabular data.',
            'Try changing it or setting --format to "jsonl".'
        ])

    loading_bar = LoadingBar(
        desc='Scraping pages',
        total=cli_args.total,
        unit='page'
    )

    worker_args = (
        cli_args.format,
        cli_args.separator
    )

    def on_irrelevant_row(reason, row, i):
        loading_bar.print('Row n°{n} could not be processed: {reason}'.format(n=i + 1, reason=reason))
        loading_bar.update()

    input_data = cli_args.report

    if cli_args.glob is not None:
        input_data = dummy_csv_file_from_glob(cli_args.glob, cli_args.input_dir)

    reader = casanova.reader(input_data)

    if (
        cli_args.glob is None and
        'raw_contents' not in reader.headers and
        not isdir(cli_args.input_dir)
    ):
        loading_bar.die([
            'Could not find the "%s" directory!' % cli_args.input_dir,
            'Did you forget to specify it with -i/--input-dir?'
        ])

    files = create_report_iterator(
        cli_args,
        reader,
        worker_args=worker_args,
        on_irrelevant_row=on_irrelevant_row
    )

    if cli_args.format == 'csv':
        output_writer = csv.DictWriter(cli_args.output, fieldnames=scraper.headers)
        output_writer.writeheader()
    else:
        output_writer = ndjson.writer(cli_args.output)

    pool = LazyPool(
        cli_args.processes,
        initializer=init_process,
        initargs=(scraper.definition, cli_args.strain)
    )

    loading_bar.update_stats(p=pool.processes)

    with pool:
        for error, items in pool.imap_unordered(worker, files):
            loading_bar.update()

            if error is not None:
                if isinstance(error, (ScraperEvalError, ScraperEvalTypeError, ScraperEvalNoneError)):
                    loading_bar.print(report_scraper_evaluation_error(error), end='')
                else:
                    loading_bar.print(error, repr(error))
                loading_bar.inc('errors')
                continue

            for item in items:
                output_writer.writerow(item)



# =============================================================================
# Minet CrowdTangle Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class CrowdTangleError(MinetError):
    pass


class CrowdTangleMissingTokenError(CrowdTangleError):
    pass


class CrowdTangleInvalidTokenError(CrowdTangleError):
    pass


class CrowdTangleInvalidRequestError(CrowdTangleError):
    def __init__(self, message=None, status=None, code=None):
        super().__init__(message)
        self.code = code
        self.status = status

    def __str__(self):
        if self.status is None:
            return super().__str__()

        return super().__str__() + ', Code: %i, Status: %i' % (self.code, self.status)


class CrowdTangleMissingStartDateError(CrowdTangleError):
    pass


class CrowdTangleInvalidJSONError(CrowdTangleError):
    pass


class CrowdTangleRateLimitExceeded(CrowdTangleError):
    pass



# =============================================================================
# Minet Google CLI Action
# =============================================================================
#
# Logic of the `google` action.
#
from minet.cli.utils import die


def google_action(cli_args):

    if cli_args.google_action == 'sheets':
        from minet.cli.google.sheets import google_sheets_action
        google_sheets_action(cli_args)



# =============================================================================
# Minet CrowdTangle Links Summary
# =============================================================================
#
# Function related to link summary using CrowdTangle's APIs.
#
from ebbe import getpath
from urllib.parse import quote

from minet.crowdtangle.exceptions import (
    CrowdTangleMissingTokenError,
    CrowdTangleInvalidTokenError,
    CrowdTangleInvalidRequestError
)
from minet.web import request_json
from minet.crowdtangle.constants import (
    CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE,
    CROWDTANGLE_SUMMARY_SORT_TYPES
)
from minet.crowdtangle.formatters import (
    format_post,
    format_summary
)

URL_TEMPLATE = (
    'https://api.crowdtangle.com/links'
    '?token=%(token)s'
    '&count=%(count)s'
    '&startDate=%(start_date)s'
    '&includeSummary=true'
    '&link=%(link)s'
    '&sortBy=%(sort_by)s'
)


def url_forge(link, token, start_date, sort_by, platforms=None, include_posts=False):

    base_url = URL_TEMPLATE % {
        'token': token,
        'count': 1 if not include_posts else 1000,
        'start_date': start_date,
        'link': quote(link, safe=''),
        'sort_by': sort_by
    }

    if platforms:
        base_url += '&platforms=%s' % ','.join(platforms)

    return base_url


def crowdtangle_summary(pool, link, token=None, start_date=None, with_top_posts=False,
                        sort_by=CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE, raw=False, platforms=None):

    if token is None:
        raise CrowdTangleMissingTokenError

    if not isinstance(start_date, str):
        raise TypeError('minet.crowdtangle.summary: expecting a `start_date` kwarg.')

    if sort_by not in CROWDTANGLE_SUMMARY_SORT_TYPES:
        raise TypeError('minet.crowdtangle.summary: unknown `sort_by`.')

    # Fetching
    api_url = url_forge(
        link,
        token,
        start_date,
        sort_by,
        platforms,
        with_top_posts
    )

    err, response, data = request_json(api_url, pool=pool)

    if err is not None:
        raise err

    if response.status == 401:
        raise CrowdTangleInvalidTokenError

    if response.status >= 400:
        raise CrowdTangleInvalidRequestError(api_url)

    stats = getpath(data, ['result', 'summary', 'facebook'])
    posts = getpath(data, ['result', 'posts']) if with_top_posts else None

    if stats is not None:
        if not raw:
            stats = format_summary(stats)

    if not with_top_posts:
        return stats

    else:
        if not raw:
            posts = [format_post(post, link=link) for post in posts]

        return stats, posts



# =============================================================================
# Minet CrowdTangle Lists CLI Action
# =============================================================================
#
# Logic of the `ct lists` action.
#
import csv

from minet.cli.utils import die
from minet.crowdtangle.constants import CROWDTANGLE_LIST_CSV_HEADERS
from minet.crowdtangle.client import CrowdTangleAPIClient
from minet.crowdtangle.exceptions import CrowdTangleInvalidTokenError


def crowdtangle_lists_action(cli_args):

    client = CrowdTangleAPIClient(cli_args.token, rate_limit=cli_args.rate_limit)
    writer = csv.writer(cli_args.output)
    writer.writerow(CROWDTANGLE_LIST_CSV_HEADERS)

    try:
        lists = client.lists()

        for l in lists:
            writer.writerow(l)

    except CrowdTangleInvalidTokenError:
        die([
            'Your API token is invalid.',
            'Check that you indicated a valid one using the `--token` argument.'
        ])



# =============================================================================
# Minet YouTube Formatters
# =============================================================================
#
# Various data formatters for YouTube data.
#
from casanova import namedrecord
from ebbe import getpath

from minet.youtube.constants import (
    YOUTUBE_VIDEO_CSV_HEADERS,
    YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS,
    YOUTUBE_COMMENT_CSV_HEADERS
)

YouTubeVideo = namedrecord(
    'YouTubeVideo',
    YOUTUBE_VIDEO_CSV_HEADERS,
    boolean=['has_caption']
)

YouTubeVideoSnippet = namedrecord(
    'YouTubeVideoSnippet',
    YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS
)

YouTubeComment = namedrecord(
    'YoutubeComment',
    YOUTUBE_COMMENT_CSV_HEADERS
)


def get_int(item, key):
    nb = item.get(key)

    if nb is not None:
        return int(nb)

    return nb


def format_video(item):
    snippet = item['snippet']
    stats = item['statistics']
    details = item['contentDetails']

    row = YouTubeVideo(
        item['id'],
        snippet['publishedAt'],
        snippet['channelId'],
        snippet['title'],
        snippet['description'],
        snippet['channelTitle'],
        get_int(stats, 'viewCount'),
        get_int(stats, 'likeCount'),
        get_int(stats, 'dislikeCount'),
        get_int(stats, 'favoriteCount'),
        get_int(stats, 'commentCount'),
        details['duration'],
        details['caption'] == 'true'
    )

    return row


def format_video_snippet(item):
    snippet = item['snippet']

    row = YouTubeVideoSnippet(
        item['id']['videoId'],
        snippet['publishedAt'],
        snippet['channelId'],
        snippet['title'],
        snippet['description'],
        snippet['channelTitle']
    )

    return row


def format_comment(item):
    meta = item['snippet']
    snippet = getpath(item, ['snippet', 'topLevelComment', 'snippet'])

    row = YouTubeComment(
        meta['videoId'],
        item['id'],
        snippet['authorDisplayName'],
        getpath(snippet, ['authorChannelId', 'value']),
        snippet['textOriginal'],
        int(snippet['likeCount']),
        snippet['publishedAt'],
        snippet['updatedAt'],
        int(meta['totalReplyCount']),
        None
    )

    return row


def format_reply(item, video_id=None):
    snippet = item['snippet']

    row = YouTubeComment(
        video_id if video_id is not None else snippet['videoId'],
        item['id'],
        snippet['authorDisplayName'],
        getpath(snippet, ['authorChannelId', 'value']),
        snippet['textOriginal'],
        int(snippet['likeCount']),
        snippet['publishedAt'],
        snippet['updatedAt'],
        None,
        snippet['parentId']
    )

    return row



# =============================================================================
# Minet CrowdTangle Formatters
# =============================================================================
#
# Various data formatters for CrowdTangle API data.
#
from casanova import namedrecord

from minet.crowdtangle.constants import (
    CROWDTANGLE_POST_TYPES,
    CROWDTANGLE_REACTION_TYPES,
    CROWDTANGLE_POST_CSV_HEADERS,
    CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK,
    CROWDTANGLE_SUMMARY_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS,
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN,
    CROWDTANGLE_LIST_CSV_HEADERS,
    CROWDTANGLE_STATISTICS,
    CROWDTANGLE_FULL_STATISTICS
)

CrowdTanglePost = namedrecord(
    'CrowdTanglePost',
    CROWDTANGLE_POST_CSV_HEADERS,
    boolean=['account_verified'],
    plural=['links', 'expanded_links'],
    json=['media']
)

CrowdTanglePostWithLink = namedrecord(
    'CrowdTanglePostWithLink',
    CROWDTANGLE_POST_CSV_HEADERS_WITH_LINK,
    boolean=['account_verified'],
    plural=['links', 'expanded_links'],
    json=['media']
)

CrowdTangleSummary = namedrecord(
    'CrowdTangleSummary',
    CROWDTANGLE_SUMMARY_CSV_HEADERS
)

CrowdTangleLeaderboard = namedrecord(
    'CrowdTangleLeaderboard',
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS,
    boolean=['verified']
)

CrowdTangleLeaderboardWithBreakdown = namedrecord(
    'CrowdTangleLeaderboardWithBreakdown',
    CROWDTANGLE_LEADERBOARD_CSV_HEADERS_WITH_BREAKDOWN,
    boolean=['verified']
)

CrowdTangleList = namedrecord(
    'CrowdTangleList',
    CROWDTANGLE_LIST_CSV_HEADERS
)


def map_key(key, target):
    return [item[key] for item in target]


def format_post(post, link=None):
    row = [
        post['id'],
        post['platformId'],
        post['platform'],
        post['type'],
        post.get('title'),
        post.get('caption'),
        post.get('message'),
        post.get('description'),
        post['date'].split(' ', 1)[0],
        post['date'],
        post['updated'],
        post.get('link'),
        post.get('postUrl'),
        post['score'],
        post.get('videoLengthMS'),
        post.get('liveVideoStatus')
    ]

    if link:
        row = [link] + row

    stats = post['statistics']
    actual_stats = stats['actual']
    expected_stats = stats['expected']

    for name in CROWDTANGLE_STATISTICS:
        key = '%sCount' % name

        row.append(actual_stats.get(key, ''))
        row.append(expected_stats.get(key, ''))

    account = post['account']

    row.extend([
        # Account
        account['id'],
        account.get('platformId'),
        account.get('platform'),
        account['name'],
        account.get('handle'),
        account.get('profileImage'),
        account['subscriberCount'],
        account['url'],
        account['verified'],
        account.get('accountType'),
        account.get('pageAdminTopCountry'),

        # Remaining
        map_key('original', post.get('expandedLinks', [])),
        map_key('expanded', post.get('expandedLinks', [])),
        post.get('media')
    ])

    if link is not None:
        return CrowdTanglePostWithLink(*row)

    return CrowdTanglePost(*row)


def format_summary(stats):
    row = (stats['%sCount' % t] for t in CROWDTANGLE_REACTION_TYPES)
    return CrowdTangleSummary(*row)


def format_leaderboard(item, with_breakdown=False):
    account = item['account']
    subscriber_data = item['subscriberData']

    row = [
        account['id'],
        account['name'],
        account.get('handle'),
        account.get('profileImage'),
        account['subscriberCount'],
        account['url'],
        account['verified'],
        subscriber_data['initialCount'],
        subscriber_data['finalCount'],
        subscriber_data.get('notes')
    ]

    summary = item['summary']

    for key, _ in CROWDTANGLE_FULL_STATISTICS:
        row.append(summary.get(key))

    if with_breakdown:
        breakdown = item['breakdown']

        for post_type in CROWDTANGLE_POST_TYPES:

            data = breakdown.get(post_type)

            for key, _ in CROWDTANGLE_FULL_STATISTICS:
                row.append(data.get(key) if data else None)

    if with_breakdown:
        return CrowdTangleLeaderboardWithBreakdown(*row)

    return CrowdTangleLeaderboard(*row)


def format_list(item):
    return CrowdTangleList(
        item['id'],
        item['title'],
        item['type']
    )



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
import casanova
from os.path import isdir
from trafilatura.core import bare_extraction

from minet.multiprocessing import LazyPool
from minet.encodings import is_supported_encoding
from minet.fs import read_potentially_gzipped_path
from minet.cli.utils import (
    create_report_iterator,
    dummy_csv_file_from_glob,
    LoadingBar
)
from minet.cli.reporters import report_error
from minet.exceptions import TrafilaturaError, UnknownEncodingError
from minet.cli.constants import DEFAULT_CONTENT_FOLDER

OUTPUT_ADDITIONAL_HEADERS = [
    'extract_error',
    'canonical_url',
    'title',
    'description',
    'raw_content',
    'comments',
    'author',
    'categories',
    'tags',
    'date',
    'sitename'
]

PADDING = [''] * (len(OUTPUT_ADDITIONAL_HEADERS) - 1)


def singular(result, key):
    return result.get(key, '') or ''


def plural(result, key):
    l = result.get(key, []) or []

    if not l:
        return ''

    items = []

    for item in l:
        for subitem in item.split(','):
            subitem = subitem.strip()

            if subitem:
                items.append(subitem)

    return '|'.join(items)


def format_trafilatura_result(result):
    return [
        '',
        singular(result, 'url'),
        singular(result, 'title'),
        singular(result, 'description'),
        singular(result, 'text'),
        singular(result, 'comments'),
        singular(result, 'author'),
        plural(result, 'categories'),
        plural(result, 'tags'),
        singular(result, 'date'),
        singular(result, 'sitename')
    ]


def format_error(error):
    return [error] + PADDING


def worker(payload):
    row, _, path, encoding, content, _ = payload

    if not is_supported_encoding(encoding):
        return UnknownEncodingError('Unknown encoding: "%s"' % encoding), row, None

    # Reading file
    if content is None:
        try:
            raw_html = read_potentially_gzipped_path(path, encoding=encoding)
        except (FileNotFoundError, UnicodeDecodeError) as e:
            return e, row, None
    else:
        raw_html = content

    # Attempting extraction
    try:
        # https://trafilatura.readthedocs.io/en/latest/corefunctions.html
        # TODO: discuss deduplication
        # TODO: fallback options
        result = bare_extraction(raw_html)
    except Exception as e:
        return TrafilaturaError(reason=e), row, None

    if result is None:
        return None, row, None

    return None, row, format_trafilatura_result(result)


def extract_action(cli_args):
    if cli_args.glob is None and cli_args.input_dir is None:
        cli_args.input_dir = DEFAULT_CONTENT_FOLDER

    input_data = cli_args.report

    if cli_args.glob is not None:
        input_data = dummy_csv_file_from_glob(cli_args.glob, cli_args.input_dir)

    enricher = casanova.enricher(
        input_data,
        cli_args.output,
        keep=cli_args.select,
        add=OUTPUT_ADDITIONAL_HEADERS
    )

    loading_bar = LoadingBar(
        desc='Extracting content',
        total=cli_args.total,
        unit='doc'
    )

    def on_irrelevant_row(reason, row, i):
        loading_bar.update()
        loading_bar.print('Row n°{n} could not be processed: {reason}'.format(n=i + 1, reason=reason))
        enricher.writerow(row, format_error(reason))

    if (
        cli_args.glob is None and
        'raw_contents' not in enricher.headers and
        not isdir(cli_args.input_dir)
    ):
        loading_bar.die([
            'Could not find the "%s" directory!' % cli_args.input_dir,
            'Did you forget to specify it with -i/--input-dir?'
        ])

    files = create_report_iterator(
        cli_args,
        enricher,
        on_irrelevant_row=on_irrelevant_row
    )

    pool = LazyPool(cli_args.processes)

    loading_bar.update_stats(p=pool.processes)

    with pool:
        for error, row, result in pool.imap_unordered(worker, files):
            loading_bar.update()

            if error is not None:
                enricher.writerow(row, format_error(report_error(error)))
                continue

            if result is None:
                enricher.writerow(row, format_error('no-result'))
                continue

            enricher.writerow(row, result)



# =============================================================================
# Minet Argparse Helpers
# =============================================================================
#
# Miscellaneous helpers related to CLI argument parsing.
#
import os
import sys
from os.path import isdir
from io import TextIOBase
from argparse import Action, ArgumentError, ArgumentTypeError
from gettext import gettext
from tqdm.contrib import DummyTqdmFile
from casanova import Resumer, CsvCellIO
from ebbe import getpath

from minet.cli.exceptions import NotResumable
from minet.cli.utils import acquire_cross_platform_stdout, was_piped_something


class SplitterType(object):
    def __init__(self, splitchar=','):
        self.splitchar = splitchar

    def __call__(self, string):
        return string.split(self.splitchar)


class ExistingDirectoryType(object):
    def __call__(self, string):
        if not isdir(string):
            raise ArgumentTypeError('Could not find the "%s" directory!' % string)


class BooleanAction(Action):
    """
    Custom argparse action to handle --no-* flags.
    Taken from: https://thisdataguy.com/2017/07/03/no-options-with-argparse-and-python/
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, cli_args, values, option_string=None):
        setattr(cli_args, self.dest, False if option_string.startswith('--no') else True)


class InputFileAction(Action):
    def __init__(self, option_strings, dest, dummy_csv_column=None,
                 dummy_csv_guard=None, dummy_csv_error='', column_dest='column',
                 nargs='?', **kwargs):

        if dummy_csv_guard is not None and not callable(dummy_csv_guard):
            raise TypeError

        self.dummy_csv_column = dummy_csv_column
        self.dummy_csv_guard = dummy_csv_guard
        self.dummy_csv_error = dummy_csv_error
        self.column_dest = column_dest

        super().__init__(
            option_strings,
            dest,
            default=None,
            nargs=nargs,
            **kwargs
        )

    def __call__(self, parser, cli_args, value, option_string=None):
        if value is None:
            f = sys.stdin

            # No stdin was piped and we have a "dummy" csv file to build
            if self.dummy_csv_column is not None and not was_piped_something():
                value = getattr(cli_args, self.column_dest)

                if self.dummy_csv_guard is not None and not self.dummy_csv_guard(value):
                    raise ArgumentError(self, self.dummy_csv_error + (' Got "%s"' % value))

                f = CsvCellIO(self.dummy_csv_column, value)
                setattr(cli_args, self.column_dest, self.dummy_csv_column)
        else:
            try:
                f = open(value, 'r', encoding='utf-8')
            except OSError as e:
                args = {'filename': value, 'error': e}
                message = gettext('can\'t open \'%(filename)s\': %(error)s')
                raise ArgumentError(self, message % args)

        setattr(cli_args, self.dest, f)


class OutputFileOpener(object):
    def __init__(self, path=None, resumer_class=None, resumer_kwargs={}, stdout_fallback=True):
        self.path = path
        self.resumer_class = resumer_class
        self.resumer_kwargs = resumer_kwargs
        self.stdout_fallback = stdout_fallback

    def open(self, cli_args, resume=False):
        if self.path is None:
            if not self.stdout_fallback:
                return None

            if resume:
                raise RuntimeError

            return DummyTqdmFile(acquire_cross_platform_stdout())

        if resume and self.resumer_class is not None:
            resumer_kwargs = self.resumer_kwargs

            if callable(self.resumer_kwargs):
                resumer_kwargs = self.resumer_kwargs(cli_args)

            return self.resumer_class(self.path, **resumer_kwargs)

        mode = 'a' if resume else 'w'

        # As per #254: newline='' is necessary for CSV output on windows to avoid
        # outputting extra lines because of a '\r\r\n' end of line...
        return open(self.path, mode, encoding='utf-8', newline='')


DEFAULT_OUTPUT_FILE_HELP = 'Path to the output file. By default, the results will be printed to stdout.'


class OutputFileAction(Action):
    def __init__(self, option_strings, dest, resumer=None, resumer_kwargs={},
                 help=DEFAULT_OUTPUT_FILE_HELP, stdout_fallback=True, **kwargs):
        self.resumer = resumer
        self.resumer_kwargs = resumer_kwargs
        self.stdout_fallback = stdout_fallback
        super().__init__(
            option_strings,
            dest,
            help=help,
            default=OutputFileOpener(
                resumer_class=resumer,
                resumer_kwargs=resumer_kwargs,
                stdout_fallback=stdout_fallback
            ),
            **kwargs
        )

    def __call__(self, parser, cli_args, value, option_string=None):
        opener = OutputFileOpener(
            value,
            resumer_class=self.resumer,
            resumer_kwargs=self.resumer_kwargs,
            stdout_fallback=self.stdout_fallback
        )
        setattr(cli_args, self.dest, opener)


def rc_key_to_env_var(key):
    return 'MINET_%s' % '_'.join(token.upper() for token in key)


class WrappedConfigValue(object):
    def __init__(self, key, default, _type):
        self.key = key
        self.default = default
        self.type = _type

    def resolve(self, config):

        # Attempting to resolve env variable
        env_var = rc_key_to_env_var(self.key)
        env_value = os.environ.get(env_var, '').strip()

        if env_value:
            return self.type(env_value)

        return getpath(config, self.key, self.default)


class ConfigAction(Action):
    def __init__(self, option_strings, dest, rc_key, default=None, **kwargs):
        if 'help' in kwargs:
            kwargs['help'] = kwargs['help'].rstrip('.') + '. Can also be configured in a .minetrc file as "%s" or read from the %s env variable.' % (
                '.'.join(rc_key),
                rc_key_to_env_var(rc_key)
            )

        super().__init__(
            option_strings,
            dest,
            default=WrappedConfigValue(
                rc_key,
                default,
                kwargs.get('type', str)
            ),
            **kwargs
        )

    def __call__(self, parser, cli_args, values, option_string=None):
        setattr(cli_args, self.dest, values)


def resolve_arg_dependencies(cli_args, config):
    to_close = []

    # Validation
    if getattr(cli_args, 'resume', False) and cli_args.output.path is None:
        raise NotResumable

    # Unwrapping values
    for name in vars(cli_args):
        value = getattr(cli_args, name)

        # Solving wrapped config values
        if isinstance(value, WrappedConfigValue):
            setattr(cli_args, name, value.resolve(config))

        # Opening output files
        if isinstance(value, OutputFileOpener):
            value = value.open(cli_args, resume=getattr(cli_args, 'resume', False))
            setattr(cli_args, name, value)

        # Finding buffers to close eventually
        if (
            (
                isinstance(value, TextIOBase) and
                value is not sys.stdin and
                value is not sys.stdout and
                value is not sys.stderr
            ) or
            isinstance(value, Resumer)
        ):
            to_close.append(value)

    return to_close



# =============================================================================
# Minet CrowdTangle Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle namespace.
#
import json
from datetime import date, datetime, timedelta

from minet.utils import rate_limited_from_state
from minet.web import request, create_request_retryer
from minet.crowdtangle.exceptions import (
    CrowdTangleMissingStartDateError,
    CrowdTangleInvalidTokenError,
    CrowdTangleInvalidRequestError,
    CrowdTangleInvalidJSONError,
    CrowdTangleRateLimitExceeded
)

DAY_DELTA = timedelta(days=1)


def day_range(end):
    day_delta = timedelta(days=1)

    start_date = date(*[int(i) for i in end.split('-')])
    current_date = date.today() + day_delta

    while start_date != current_date:
        end_date = current_date
        current_date -= day_delta

        yield current_date.isoformat(), end_date.isoformat()


def years_iter(start_date, end_date):
    current_year = int(end_date[:4])
    target_year = int(start_date[:4])

    current_start_date = str(current_year) + '-01-01T00:00:00'
    current_end_date = end_date

    while current_year > target_year:
        yield current_start_date, current_end_date

        current_year -= 1
        current_end_date = str(current_year) + '-12-31T23:59:59'
        current_start_date = str(current_year) + current_start_date[4:]

    yield start_date, current_end_date


def infer_end_date():
    return datetime.now().isoformat(timespec='seconds')


def get_last_day_of_month(year, month):
    d = 31

    while d > 27:
        try:
            attempt = date(int(year), int(month), d)
            return attempt.isoformat()[8:10]
        except ValueError:
            d -= 1

    raise TypeError('could not find last day of month')


def complement_date(d, bound):
    if len(d) == 4:
        d += '-'
        d += '01' if bound == 'start' else '12'

    if len(d) == 7:
        d += '-'
        d += '01' if bound == 'start'else get_last_day_of_month(d[:4], d[5:7])

    if len(d) == 10:
        d += 'T00:00:00' if bound == 'start' else 'T23:59:59'

    return d


def step(pool, url, item_key):
    err, result = request(url, pool=pool)

    # Debug
    if err:
        raise err

    # Bad auth
    if result.status == 401:
        raise CrowdTangleInvalidTokenError

    elif result.status == 429:
        raise CrowdTangleRateLimitExceeded

    # Bad params
    if result.status >= 400:
        data = result.data.decode('utf-8')

        try:
            data = json.loads(data)
        except:
            raise CrowdTangleInvalidRequestError(data)

        raise CrowdTangleInvalidRequestError(data['message'], code=data['code'], status=result.status)

    try:
        data = json.loads(result.data)['result']
    except (json.decoder.JSONDecodeError, TypeError, KeyError):
        raise CrowdTangleInvalidJSONError

    items = None

    if item_key in data:
        items = data[item_key]

        if len(items) == 0:
            items = None

    # Extracting next link
    pagination = data['pagination']
    next_page = pagination['nextPage'] if 'nextPage' in pagination else None

    return items, next_page


def default_item_id_getter(item):
    return item['id']


def make_paginated_iterator(url_forge, item_key, formatter,
                            item_id_getter=default_item_id_getter):

    def create_iterator(pool, token, rate_limiter_state, limit=None,
                        raw=False, per_call=False, detailed=False,
                        namespace=None, before_sleep=None, **kwargs):

        if namespace is not None:
            kwargs = vars(namespace)
        else:
            kwargs['token'] = token

        # Checking we have the necessary dates
        if kwargs.get('sort_by') == 'date':
            if kwargs.get('start_date') is None:
                raise CrowdTangleMissingStartDateError

            # Inferring end date to be now, this will be important later
            if kwargs.get('end_date') is None:
                kwargs['end_date'] = infer_end_date()

        # Complementing dates
        if kwargs.get('start_date') is not None:
            kwargs['start_date'] = complement_date(kwargs['start_date'], 'start')

        if kwargs.get('end_date') is not None:
            kwargs['end_date'] = complement_date(kwargs['end_date'], 'end')

        N = 0
        C = 0
        last_url = None
        last_items = set()

        has_limit = limit is not None

        rate_limited_step = rate_limited_from_state(rate_limiter_state)(step)

        retryer = create_request_retryer(
            additional_exceptions=[CrowdTangleRateLimitExceeded, CrowdTangleInvalidJSONError],
            before_sleep=before_sleep
        )

        # Chunking
        need_to_chunk = kwargs.get('sort_by') == 'date'
        chunk_size = kwargs.get('chunk_size', 500)
        current_chunk_size = 0
        shifts = 0
        years = years_iter(kwargs['start_date'], kwargs['end_date']) if need_to_chunk else None

        def rotate_year():
            try:
                start_date, end_date = next(years)
                kwargs['start_date'] = start_date
                kwargs['end_date'] = end_date
            except StopIteration:
                return False

            return True

        if need_to_chunk:
            rotate_year()

        # Starting url
        url = url_forge(**kwargs)

        while True:
            C += 1

            items, next_url = retryer(rate_limited_step, pool, url, item_key)

            # We have exhausted the available data
            if items is None:

                if need_to_chunk:
                    could_rotate = rotate_year()

                    if could_rotate:
                        url = url_forge(**kwargs)
                        continue

                break

            enough_to_stop = False
            n = 0

            last_url = url

            acc = []

            for item in items:

                # Avoiding duplicating items due to race conditions
                if item_id_getter(item) in last_items:
                    continue

                current_date = item.get('date')

                n += 1
                N += 1
                current_chunk_size += 1

                if not raw:
                    item = formatter(item)

                acc.append(item)

                if has_limit and N >= limit:
                    enough_to_stop = True
                    break

            if per_call:
                if detailed:
                    details = None

                    if need_to_chunk:
                        details = {
                            'date': current_date,
                            'shifts': shifts
                        }

                    yield details, acc
                else:
                    yield acc
            else:
                yield from acc

            if enough_to_stop:
                break

            # We need to track last items to avoid registering the same one twice
            last_items = set(item_id_getter(item) for item in items)

            # Paginating
            if next_url is None:
                if need_to_chunk:
                    could_rotate = rotate_year()

                    if could_rotate:
                        url = url_forge(**kwargs)
                        continue

                break

            # Handling chunking
            if need_to_chunk and current_chunk_size >= chunk_size:
                current_chunk_size = 0
                shifts += 1
                kwargs['end_date'] = items[-1]['date'].replace(' ', 'T')
                url = url_forge(**kwargs)
                continue

            url = next_url

    return create_iterator



# =============================================================================
# Minet Scraper Utils
# =============================================================================
#
# Miscellaneous helper functions used throughout the minet.scrape package.
#
from functools import partial

from minet.scrape.constants import (
    SELECT_ALIASES,
    ITERATOR_ALIASES
)


def get_aliases(aliases, target, with_key=False):
    for alias in aliases:
        if alias in target:
            if with_key:
                return alias, target[alias]

            return target[alias]

    if with_key:
        return None, None

    return None


get_sel = partial(get_aliases, SELECT_ALIASES)
get_iterator = partial(get_aliases, ITERATOR_ALIASES)



# =============================================================================
# Minet Scrapers Standard Library
# =============================================================================
#
# Collection of helpers, and depencies exposed to scraper evaluation.
#
import re
import json
import soupsieve
from urllib.parse import urljoin
from bs4 import Tag, NavigableString

from minet.utils import parse_date
from minet.scrape.constants import BLOCK_ELEMENTS, CONTENT_BLOCK_ELEMENTS


LEADING_WHITESPACE_RE = re.compile(r'^\s')
TRAILING_WHITESPACE_RE = re.compile(r'\s$')
LINE_STRIPPER_RE = re.compile(r'\n +| +\n')
PARAGRAPH_NORMALIZER_RE = re.compile(r'\n{3,}')
SPACE_SQUEEZER_RE = re.compile(r' {2,}')
WHITESPACE_SQUEEZER_RE = re.compile(r'\s{2,}')


def has_leading_whitespace(string):
    return LEADING_WHITESPACE_RE.match(string) is not None


def has_trailing_whitespace(string):
    return TRAILING_WHITESPACE_RE.search(string) is not None


def is_block_element(element):
    if isinstance(element, NavigableString):
        return False

    return (
        element.name in BLOCK_ELEMENTS or
        element.name == 'html' or
        element.name == '[document]'
    )


def is_inline_element(element):
    return not is_block_element(element)


def get_element_display(element):
    return 'block' if is_block_element(element) else 'inline'


# NOTE: could be optimized?
def get_block_parent(element):
    parent = element.parent

    while True:
        if parent is not None and is_block_element(parent):
            return parent

        parent = parent.parent


def get_previous_sibling(element):
    while element.previous_sibling is not None:
        if isinstance(element.previous_sibling, Tag):
            return element.previous_sibling

        element = element.previous_sibling

    return None


def get_display_text(element):
    if isinstance(element, list):
        pieces = (get_display_text(el) for el in element)

        return '\n\n'.join(piece for piece in pieces if piece.strip())

    def accumulator():
        previous_block_parent = None
        last_string = None

        for descendant in element.descendants:
            if not isinstance(descendant, NavigableString):

                if descendant.name == 'br':
                    yield '\n'

                elif descendant.name == 'hr':
                    yield '\n\n'

                elif descendant.name in CONTENT_BLOCK_ELEMENTS:
                    yield '\n'

                else:
                    sibling = get_previous_sibling(descendant)

                    if sibling:
                        if sibling.name in CONTENT_BLOCK_ELEMENTS:
                            yield '\n'
                        else:
                            yield ' '

                continue

            if descendant.parent.name == 'pre':
                yield '\n' + str(descendant)
                continue

            string = WHITESPACE_SQUEEZER_RE.sub(' ', descendant.strip('\n'))

            if not string:
                continue

            block_parent = get_block_parent(descendant)

            if block_parent is not previous_block_parent:
                previous_block_parent = block_parent
                yield '\n'

            if last_string and last_string.endswith(' '):
                string = string.lstrip()

            if string:
                yield string

            last_string = string

    result = ''.join(accumulator())
    result = LINE_STRIPPER_RE.sub('\n', result)
    result = PARAGRAPH_NORMALIZER_RE.sub('\n\n', result)
    result = SPACE_SQUEEZER_RE.sub(' ', result)
    result = result.strip()

    return result


def get_default_evaluation_context():
    return {

        # Dependencies
        'json': json,
        'urljoin': urljoin,
        're': re,
        'soupsieve': soupsieve,

        # Helpers
        'get_display_text': get_display_text,
        'parse_date': parse_date
    }



# =============================================================================
# Minet Facebook Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class FacebookError(MinetError):
    pass


class FacebookInvalidCookieError(FacebookError):
    pass


class FacebookInvalidTargetError(FacebookError):
    pass



# =============================================================================
# Minet Youtube CLI Action
# =============================================================================
#
# Logic of the `yt` action.
#
from minet.cli.utils import die


def check_key(cli_args):

    # A key is required to used the API
    if not cli_args.key:
        die([
            'A key is required to access YouTube API.',
            'You can provide it using the --key argument.'
        ])


def youtube_action(cli_args):

    if cli_args.yt_action == 'videos':
        check_key(cli_args)

        from minet.cli.youtube.videos import videos_action
        videos_action(cli_args)

    elif cli_args.yt_action == 'comments':
        check_key(cli_args)

        from minet.cli.youtube.comments import comments_action
        comments_action(cli_args)

    elif cli_args.yt_action == 'captions':
        from minet.cli.youtube.captions import captions_action
        captions_action(cli_args)

    elif cli_args.yt_action == 'search':
        check_key(cli_args)

        from minet.cli.youtube.search import search_action
        search_action(cli_args)



# =============================================================================
# Minet Scraper Constants
# =============================================================================
#
# Constants used throughout the minet.scrape package.
#
EXTRACTOR_NAMES = ['text', 'display_text', 'html', 'inner_html', 'outer_html']
SELECT_ALIASES = ['sel', '$']
ITERATOR_ALIASES = ['iterator', '$$']
PLURAL_MODIFIERS = ['filter', 'filter_eval', 'uniq']
BURROWING_KEYS = ITERATOR_ALIASES + ['item', 'fields']
LEAF_KEYS = ['attr', 'extract', 'get_context', 'eval', 'default']

BLOCK_ELEMENTS = {
    'article',
    'aside',
    'blockquote',
    'body',
    'br',
    'button',
    'canvas',
    'caption',
    'col',
    'colgroup',
    'dd',
    'div',
    'dl',
    'dt',
    'embed',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'header',
    'hgroup',
    'hr',
    'li',
    'map',
    'object',
    'ol',
    'output',
    'p',
    'pre',
    'progress',
    'section',
    'table',
    'tbody',
    'textarea',
    'tfoot',
    'thead',
    'tr',
    'ul',
    'video'
}

CONTENT_BLOCK_ELEMENTS = {
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'p',
    'ul',
    'ol',
    'pre',
    'blockquote'
}



# =============================================================================
# Minet Mediacloud CLI Action
# =============================================================================
#
# Logic of the `mc` action.
#
from minet.cli.utils import die


def mediacloud_action(cli_args):

    # A token is needed to be able to access the API
    if not cli_args.token:
        die([
            'A token is needed to be able to access Mediacloud\'s API.',
            'You can provide one using the `--token` argument.'
        ])

    if cli_args.mc_action == 'medias':
        from minet.cli.mediacloud.medias import mediacloud_medias_action
        mediacloud_medias_action(cli_args)

    if cli_args.mc_action == 'topic':
        from minet.cli.mediacloud.topic import mediacloud_topic_action
        mediacloud_topic_action(cli_args)

    elif cli_args.mc_action == 'search':
        from minet.cli.mediacloud.search import mediacloud_search_action
        mediacloud_search_action(cli_args)



# =============================================================================
# Minet YouTube Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class YouTubeError(MinetError):
    pass


class YouTubeInvalidAPIKeyError(YouTubeError):
    pass


class YouTubeInvalidAPICall(YouTubeError):
    def __init__(self, url, status, data):
        super().__init__()
        self.url = url
        self.status = status
        self.data = data

    def __str__(self):
        return super().__str__() + ', Url: %s, Status: %i, Data: %s' % (self.url, self.status, self.data)


class YouTubeInvalidVideoId(YouTubeError):
    pass



# =============================================================================
# Minet Mediacloud Topic CLI Action
# =============================================================================
#
# Logic of the `mc topic` action.
#
import csv

from minet.cli.utils import LoadingBar
from minet.mediacloud import MediacloudAPIClient
from minet.mediacloud.constants import MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS


def mediacloud_topic_action(cli_args):
    writer = csv.writer(cli_args.output)
    writer.writerow(MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS)

    loading_bar = LoadingBar(
        desc='Fetching stories',
        unit='story',
        unit_plural='stories'
    )

    client = MediacloudAPIClient(cli_args.token)

    iterator = client.topic_stories(
        cli_args.topic_id,
        media_id=cli_args.media_id,
        from_media_id=cli_args.from_media_id
    )

    for story in iterator:
        writer.writerow(story.as_csv_row())
        loading_bar.update()



# =============================================================================
# Minet Twitter Scrape CLI Action
# =============================================================================
#
# Logic of the `tw scrape` action.
#
import casanova
from twitwi.constants import TWEET_FIELDS
from twitwi import format_tweet_as_csv_row

from minet.utils import prettyprint_seconds, PseudoFStringFormatter
from minet.cli.utils import LoadingBar
from minet.twitter import TwitterAPIScraper
from minet.twitter.exceptions import (
    TwitterPublicAPIRateLimitError,
    TwitterPublicAPIOverCapacityError
)
from minet.twitter.constants import ADDITIONAL_TWEET_FIELDS

CUSTOM_FORMATTER = PseudoFStringFormatter()


def format_meta_row(meta):
    if meta is None:
        return [''] * len(ADDITIONAL_TWEET_FIELDS)

    return [
        meta.get('intervention_type'),
        meta.get('intervention_text'),
        meta.get('intervention_url')
    ]


def twitter_scrape_action(cli_args):
    scraper = TwitterAPIScraper()

    # Stats
    loading_bar = LoadingBar(
        'Collecting tweets',
        total=cli_args.limit,
        unit='tweet',
        stats={'tokens': 1, 'queries': 0}
    )

    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        add=TWEET_FIELDS + ADDITIONAL_TWEET_FIELDS,
        keep=cli_args.select
    )

    def before_sleep(retry_state):
        exc = retry_state.outcome.exception()

        if isinstance(exc, TwitterPublicAPIRateLimitError):
            loading_bar.inc('tokens')

        else:
            loading_bar.inc('failures')
            loading_bar.print(
                'Failed to call Twitter search. Will retry in %s' % prettyprint_seconds(retry_state.idle_for)
            )

    for row, query in enricher.cells(cli_args.query, with_rows=True):

        # Templating?
        if cli_args.query_template is not None:
            query = CUSTOM_FORMATTER.format(
                cli_args.query_template,
                value=query
            )

        loading_bar.print('Searching for "%s"' % query)
        loading_bar.inc('queries')

        iterator = scraper.search(
            query,
            limit=cli_args.limit,
            before_sleep=before_sleep,
            include_referenced_tweets=cli_args.include_refs,
            with_meta=True
        )

        try:
            for tweet, meta in iterator:
                loading_bar.update()

                tweet_row = format_tweet_as_csv_row(tweet)
                enricher.writerow(row, tweet_row + format_meta_row(meta))
        except TwitterPublicAPIOverCapacityError:
            loading_bar.die('Got an "Over Capacity" error. Shutting down...')



# =============================================================================
# Minet Facebook Url Likes CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving approximate
# likes count by scraping Facebook's like button plugin.
#
import re
import casanova
from urllib.parse import quote
from ural import is_url

from minet.utils import rate_limited
from minet.web import request
from minet.cli.constants import DEFAULT_PREBUFFER_BYTES
from minet.cli.utils import die, LoadingBar

REPORT_HEADERS = ['approx_likes', 'approx_likes_int']
ONE_LIKE_RE = re.compile(rb'>\s*One person likes this\.', re.I)
LIKES_RE = re.compile(rb'>\s*([\d.KM]+)\s+people\s+like', re.I)


def forge_url(url):
    return 'https://www.facebook.com/plugins/like.php?href=%s' % quote(url)


@rate_limited(5)
def make_request(url):
    err, response = request(forge_url(url), headers={'Accept-Language': 'en'})

    if response.status == 404:
        return 'not-found', None

    if err:
        return 'http-error', None

    return err, response.data


def parse_approx_likes(approx_likes, unit='K'):
    multiplier = 1000

    if unit == 'M':
        multiplier = 1000000

    return str(int(float(approx_likes[:-1]) * multiplier))


def scrape(data):

    if ONE_LIKE_RE.search(data):
        return ['1', '1']

    match = LIKES_RE.search(data)

    if match is None:
        return ['', '']

    approx_likes = match.group(1).decode()
    approx_likes_int = approx_likes

    if 'K' in approx_likes:
        approx_likes_int = parse_approx_likes(approx_likes, unit='K')

    elif 'M' in approx_likes:
        approx_likes_int = parse_approx_likes(approx_likes, unit='M')

    return [approx_likes, approx_likes_int]


def facebook_url_likes_action(cli_args):
    enricher = casanova.enricher(
        cli_args.file,
        cli_args.output,
        keep=cli_args.select,
        add=REPORT_HEADERS,
        total=cli_args.total,
        prebuffer_bytes=DEFAULT_PREBUFFER_BYTES
    )

    if cli_args.column not in enricher.headers:
        die([
            'Could not find the "%s" column containing the urls in the given CSV file.' % cli_args.column
        ])

    loading_bar = LoadingBar(
        desc='Retrieving likes',
        unit='url',
        total=enricher.total
    )

    for row, url in enricher.cells(cli_args.column, with_rows=True):
        loading_bar.update()

        url = url.strip()

        if not url or not is_url(url, require_protocol=False):
            enricher.writerow(row)
            continue

        err, html = make_request(url)

        if err is not None:
            loading_bar.die('An error occurred while fetching like button for this url: %s' % url)

        scraped = scrape(html)

        if scraped is None:
            loading_bar.die('Could not extract Facebook likes from this url\'s like button: %s' % url)

        enricher.writerow(row, scraped)



# =============================================================================
# Minet Hyphe API Client
# =============================================================================
#
# Client for the Hyphe API.
#
from minet.web import create_pool, request_jsonrpc
from minet.hyphe.constants import (
    WEBENTITY_STATUSES,
    DEFAULT_PAGINATION_COUNT
)
from minet.hyphe.exceptions import (
    HypheCouldNotStartCorpusError,
    HypheJSONRPCError,
    HypheRequestFailError,
    HypheCorpusAuthenticationError
)


class HypheAPIClient(object):
    def __init__(self, endpoint):
        if not endpoint.endswith('/'):
            endpoint += '/'

        self.endpoint = endpoint
        self.pool = create_pool()

    def call(self, method, *args, **kwargs):
        err, result = request_jsonrpc(
            self.endpoint,
            method,
            pool=self.pool,
            *args,
            **kwargs
        )

        if err:
            return err, None

        if 'fault' in result:
            return HypheJSONRPCError(result), None

        if (
            isinstance(result, list) and
            len(result) > 0 and
            isinstance(result[0], dict) and
            result[0].get('code') == 'fail'
        ):
            msg = result[0].get('message', '')

            if 'password' in msg:
                raise HypheCorpusAuthenticationError

            raise HypheRequestFailError(msg)

        return None, result[0]

    def corpus(self, corpus, password=None):
        return HypheAPIClientCorpus(self, corpus, password)


class HypheAPIClientCorpus(object):
    def __init__(self, client, corpus, password=None):
        self.client = client
        self.corpus = corpus
        self.password = password

    def call(self, method, *args, **kwargs):
        kwargs['corpus'] = self.corpus

        return self.client.call(method, *args, **kwargs)

    def ping_until_ready(self, max_attempts=10):
        for _ in range(max_attempts):
            err, result = self.call('ping', timeout=5)

            if result['result'] == 'pong':
                return True

        raise HypheCouldNotStartCorpusError

    def ensure_is_started(self):
        err, result = self.call('start_corpus', password=self.password)

        if err:
            raise err

        # Corpus is already started
        if result['result']['status'] == 'ready':
            return True

        # Pinging until ready
        return self.ping_until_ready()

    def status(self):
        err, stats = self.call('get_status')

        if err:
            raise err

        return stats

    def count(self, statuses=None):
        if not statuses:
            statuses = WEBENTITY_STATUSES

        stats = self.status()

        webentity_counts = stats['result']['corpus']['traph']['webentities']

        total = 0

        for status in statuses:
            total += webentity_counts[status]

        info = {
            'webentities': total,
            'pages': stats['result']['corpus']['crawler']['pages_found']
        }

        return info

    def webentities(self, statuses=None):
        if not statuses:
            statuses = WEBENTITY_STATUSES

        for status in statuses:
            token = None
            next_page = None

            while True:
                if token is None:
                    err, result = self.call(
                        'store.get_webentities_by_status',
                        status=status,
                        count=DEFAULT_PAGINATION_COUNT
                    )
                else:
                    err, result = self.call(
                        'store.get_webentities_page',
                        pagination_token=token,
                        n_page=next_page
                    )

                if err:
                    raise err

                result = result['result']

                for webentity in result['webentities']:
                    yield webentity

                if 'next_page' in result and result['next_page']:
                    token = result['token']
                    next_page = result['next_page']
                else:
                    break

    def webentity_pages(self, webentity_id, include_body=False):
        token = None

        while True:
            err, result = self.call(
                'store.paginate_webentity_pages',
                webentity_id=webentity_id,
                count=DEFAULT_PAGINATION_COUNT,
                pagination_token=token,
                include_page_metas=True,
                include_page_body=include_body
            )

            if err:
                raise err

            result = result['result']

            for page in result['pages']:
                yield page

            if 'token' in result and result['token']:
                token = result['token']
            else:
                break



# =============================================================================
# Minet Mediacloud Search
# =============================================================================
#
# Function related to stories searching.
#
from urllib.parse import quote_plus

from minet.web import request_json
from minet.mediacloud.constants import (
    MEDIACLOUD_API_BASE_URL,
    MEDIACLOUD_DEFAULT_BATCH
)
from minet.mediacloud.exceptions import (
    MediacloudServerError
)
from minet.mediacloud.formatters import format_story
from minet.mediacloud.utils import get_last_processed_stories_id


def create_plural_query_component(field, values):
    query = ' AND ('
    query += ' OR '.join(('%s:%s' % (field, str(value))) for value in values)
    query += ')'

    return query


def pad_date(string):
    if len(string) == 4:
        return string + '-01-01T00:00:00Z'

    if len(string) == 7:
        return string + '-01T00:00:00Z'

    if 'T' not in string:
        return string + 'T00:00:00Z'

    return string


def query_additions(query, collections=None, medias=None,
                    publish_day=None, publish_month=None, publish_year=None):

    if collections is not None:
        query += create_plural_query_component('tags_id_media', collections)

    if medias is not None:
        query += create_plural_query_component('media_id', medias)

    if publish_day is not None:
        query += ' AND publish_day:"%s"' % pad_date(publish_day)

    if publish_month is not None:
        query += ' AND publish_month:"%s"' % pad_date(publish_month)

    if publish_year is not None:
        query += ' AND publish_year:"%s"' % pad_date(publish_year)

    # NOTE: range queries can be made with [from_date TO to_date] but must appear in filter

    return query


def url_forge(token, query, filter_query=None, collections=None, medias=None,
              publish_day=None, publish_month=None, publish_year=None,
              count=False, last_processed_stories_id=None):

    url = '%s/stories_public/%s?key=%s' % (
        MEDIACLOUD_API_BASE_URL,
        'count' if count else 'list',
        token
    )

    query = query_additions(
        query,
        collections=collections,
        medias=medias,
        publish_day=publish_day,
        publish_month=publish_month,
        publish_year=publish_year
    )

    url += '&q=%s' % quote_plus(query)

    if filter_query is not None:
        url += '&fq=%s' % quote_plus(filter_query)

    if not count:
        url += '&rows=%i' % MEDIACLOUD_DEFAULT_BATCH

        if last_processed_stories_id is not None:
            url += '&last_processed_stories_id=%i' % last_processed_stories_id

    return url


def mediacloud_search(pool, token, query, filter_query=None, count=False,
                      collections=None, medias=None, publish_day=None,
                      publish_month=None, publish_year=None, raw=False):

    def generator():
        last_processed_stories_id = None

        while True:
            url = url_forge(
                token,
                query,
                filter_query=filter_query,
                collections=collections,
                medias=medias,
                publish_day=publish_day,
                publish_month=publish_month,
                publish_year=publish_year,
                count=count,
                last_processed_stories_id=last_processed_stories_id
            )

            err, response, data = request_json(url, pool=pool)

            if err:
                raise err

            if response.status >= 500:
                raise MediacloudServerError(server_error=data.get('error'))

            if count:
                yield data['count']
                return

            for story in data:
                if not raw:
                    story = format_story(story)

                yield story

            last_processed_stories_id = get_last_processed_stories_id(data)

            if last_processed_stories_id is None:
                return

    if count:
        return next(generator())

    return generator()



# =============================================================================
# Minet Twitter Constants
# =============================================================================
#
# General constants used throughout the Twitter functions.
#
from urllib3 import Timeout


TWITTER_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
TWITTER_PUBLIC_API_AUTH_HEADER = 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'

TWITTER_API_MAX_STATUSES_COUNT = 200

ADDITIONAL_TWEET_FIELDS = [
    'intervention_type',
    'intervention_text',
    'intervention_url'
]



# =============================================================================
# Minet Instagram API Scraper
# =============================================================================
#
# Instagram public API "scraper".
#
import json
from urllib.parse import quote
from ebbe import getpath

from minet.constants import COOKIE_BROWSERS
from minet.web import create_pool, request_json, grab_cookies
from minet.instagram.constants import (
    INSTAGRAM_URL,
    INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT
)
from minet.instagram.exceptions import (
    InstagramPublicAPIInvalidResponseError,
    InstagramInvalidCookieError
)

INSTAGRAM_GRAPHQL_ENDPOINT = 'https://www.instagram.com/graphql/query/'
INSTAGRAM_HASHTAG_QUERY_HASH = '9b498c08113f1e09617a1703c22b2f32'


def forge_hashtag_search_url(name, cursor=None, count=50):
    params = {
        'tag_name': name,
        'first': count
    }

    if cursor is not None:
        params['after'] = cursor

    url = INSTAGRAM_GRAPHQL_ENDPOINT + '?query_hash=%s&variables=%s' % (
        INSTAGRAM_HASHTAG_QUERY_HASH,
        quote(json.dumps(params))
    )

    return url


class InstagramAPIScraper(object):
    def __init__(self, cookie='firefox'):
        self.pool = create_pool(timeout=INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT)

        if cookie in COOKIE_BROWSERS:
            get_cookie_for_url = grab_cookies(cookie)
            cookie = get_cookie_for_url(INSTAGRAM_URL)

        if not cookie:
            raise InstagramInvalidCookieError

        self.cookie = cookie

    def request_json(self, url):
        err, response, data = request_json(
            url,
            pool=self.pool,
            spoof_ua=True,
            headers={'Cookie': self.cookie}
        )

        if err:
            raise err

        if response.status >= 400:
            raise InstagramPublicAPIInvalidResponseError

        return data

    def search_hashtag(self, name):
        name = name.lstrip('#')
        cursor = None

        while True:
            url = forge_hashtag_search_url(name, cursor=cursor)
            print(url, cursor)

            data = self.request_json(url)

            data = getpath(data, ['data', 'hashtag', 'edge_hashtag_to_media'])
            edges = data.get('edges')

            for edge in edges:
                yield edge['node']['shortcode']

            print('Found %i posts' % len(edges))

            has_next_page = getpath(data, ['page_info', 'has_next_page'])

            if not has_next_page:
                break

            cursor = getpath(data, ['page_info', 'end_cursor'])



# =============================================================================
# Minet Multithreaded Fetch
# =============================================================================
#
# Exposing a specialized quenouille wrapper grabbing various urls from the
# web in a multithreaded fashion.
#
from collections import namedtuple
from quenouille import ThreadPoolExecutor
from ural import get_domain_name, ensure_protocol

from minet.web import (
    create_pool,
    request,
    resolve,
    extract_response_meta
)
from minet.heuristics import should_spoof_ua_when_resolving
from minet.constants import (
    DEFAULT_DOMAIN_PARALLELISM,
    DEFAULT_IMAP_BUFFER_SIZE,
    DEFAULT_THROTTLE,
    DEFAULT_URLLIB3_TIMEOUT,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_RESOLVE_MAX_REDIRECTS
)

FetchWorkerPayload = namedtuple(
    'FetchWorkerPayload',
    [
        'item',
        'domain',
        'url'
    ]
)


class FetchResult(object):
    __slots__ = ('item', 'domain', 'url', 'error', 'response', 'meta')

    def __init__(self, item, domain, url):
        self.item = item
        self.domain = domain
        self.url = url
        self.error = None
        self.response = None
        self.meta = {}

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return '<{name} empty!>'.format(name=name)

        return '<{name}{errored} url={url!r} status={status!r} ext={ext!r} encoding={encoding!r}>'.format(
            name=name,
            url=self.url,
            status=self.response.status if self.response else None,
            ext=self.meta.get('ext'),
            encoding=self.meta.get('encoding'),
            errored=' errored!' if self.error else ''
        )

    @property
    def resolved(self):
        if self.response is None:
            return None

        return self.response.geturl()


class ResolveResult(object):
    __slots__ = ('item', 'domain', 'url', 'error', 'stack')

    def __init__(self, item, domain, url):
        self.item = item
        self.domain = domain
        self.url = url
        self.error = None
        self.stack = None

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return '<{name} empty!>'.format(name=name)

        return '<{name}{errored} url={url!r} status={status!r} redirects={redirects!r}>'.format(
            name=name,
            url=self.url,
            status=self.stack[-1].status if self.stack else None,
            redirects=(len(self.stack) - 1) if self.stack else 0,
            errored=' errored!' if self.error else ''
        )


def key_by_domain_name(payload):
    return payload.domain


def payloads_iter(iterator, key=None):
    for item in iterator:
        url = item if key is None else key(item)

        if not url:
            yield FetchWorkerPayload(
                item=item,
                domain=None,
                url=None
            )

            continue

        # Url cleanup
        url = ensure_protocol(url.strip())

        yield FetchWorkerPayload(
            item=item,
            domain=get_domain_name(url),
            url=url
        )


class FetchWorker(object):
    def __init__(self, pool, *, request_args=None, max_redirects=DEFAULT_FETCH_MAX_REDIRECTS,
                 callback=None):
        self.pool = pool
        self.request_args = request_args
        self.max_redirects = max_redirects
        self.callback = callback

    def __call__(self, payload):
        item, domain, url = payload

        result = FetchResult(*payload)

        if url is None:
            return result

        # NOTE: request_args must be threadsafe
        kwargs = {}

        if self.request_args is not None:
            kwargs = self.request_args(domain, url, item)

        error, response = request(
            url,
            pool=self.pool,
            max_redirects=self.max_redirects,
            **kwargs
        )

        if error:
            result.error = error
        else:

            # Forcing urllib3 to read data in thread
            # TODO: this is probably useless and should be replaced by preload_content at the right place
            data = response.data

            # Meta
            meta = extract_response_meta(response)

            result.response = response
            result.meta = meta

            if self.callback is not None:
                self.callback(result)

        return result


class ResolveWorker(object):
    def __init__(self, pool, *, resolve_args=None, max_redirects=DEFAULT_RESOLVE_MAX_REDIRECTS,
                 follow_refresh_header=True, follow_meta_refresh=False, follow_js_relocation=False,
                 infer_redirection=False):

        self.pool = pool
        self.resolve_args = resolve_args
        self.max_redirects = max_redirects
        self.follow_refresh_header = follow_refresh_header
        self.follow_meta_refresh = follow_meta_refresh
        self.follow_js_relocation = follow_js_relocation
        self.infer_redirection = infer_redirection

    def __call__(self, payload):
        item, domain, url = payload

        result = ResolveResult(*payload)

        if url is None:
            return result

        # NOTE: resolve_args must be threadsafe
        kwargs = {}

        if self.resolve_args is not None:
            kwargs = self.resolve_args(domain, url, item)

        # NOTE: should it be just in the CLI?
        if 'spoof_ua' not in kwargs:
            kwargs['spoof_ua'] = should_spoof_ua_when_resolving(domain)

        error, stack = resolve(
            url,
            pool=self.pool,
            max_redirects=self.max_redirects,
            follow_refresh_header=self.follow_refresh_header,
            follow_meta_refresh=self.follow_meta_refresh,
            follow_js_relocation=self.follow_js_relocation,
            infer_redirection=self.infer_redirection,
            **kwargs
        )

        result.error = error
        result.stack = stack

        return result


class HTTPThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(self, max_workers=None, insecure=False, timeout=DEFAULT_URLLIB3_TIMEOUT,
                 **kwargs):
        super().__init__(max_workers, **kwargs)
        self.pool = create_pool(threads=max_workers, insecure=insecure, timeout=timeout)

    def imap(self, *args, **kwargs):
        raise NotImplementedError


class FetchThreadPoolExecutor(HTTPThreadPoolExecutor):
    def imap_unordered(self, iterator, *, key=None, throttle=DEFAULT_THROTTLE, request_args=None,
                       buffer_size=DEFAULT_IMAP_BUFFER_SIZE, domain_parallelism=DEFAULT_DOMAIN_PARALLELISM,
                       max_redirects=DEFAULT_FETCH_MAX_REDIRECTS, callback=None):

        # TODO: validate
        iterator = payloads_iter(iterator, key=key)
        worker = FetchWorker(
            self.pool,
            request_args=request_args,
            max_redirects=max_redirects,
            callback=callback
        )

        return super().imap_unordered(
            iterator,
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle
        )


class ResolveThreadPoolExecutor(HTTPThreadPoolExecutor):
    def imap_unordered(self, iterator, *, key=None, throttle=DEFAULT_THROTTLE, resolve_args=None,
                       buffer_size=DEFAULT_IMAP_BUFFER_SIZE, domain_parallelism=DEFAULT_DOMAIN_PARALLELISM,
                       max_redirects=DEFAULT_FETCH_MAX_REDIRECTS, follow_refresh_header=True,
                       follow_meta_refresh=False, follow_js_relocation=False, infer_redirection=False):

        # TODO: validate
        iterator = payloads_iter(iterator, key=key)
        worker = ResolveWorker(
            self.pool,
            resolve_args=resolve_args,
            max_redirects=max_redirects,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            infer_redirection=infer_redirection
        )

        return super().imap_unordered(
            iterator,
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle
        )


def multithreaded_fetch(iterator, key=None, request_args=None, threads=25,
                        throttle=DEFAULT_THROTTLE, buffer_size=DEFAULT_IMAP_BUFFER_SIZE,
                        insecure=False, timeout=DEFAULT_URLLIB3_TIMEOUT,
                        domain_parallelism=DEFAULT_DOMAIN_PARALLELISM,
                        max_redirects=5, wait=True, daemonic=False, callback=None):
    """
    Function returning a multithreaded iterator over fetched urls.

    Args:
        iterator (iterable): An iterator over urls or arbitrary items.
        key (callable, optional): Function extracting url from yielded items.
        request_args (callable, optional): Function returning specific
            arguments to pass to the request util per yielded item.
        threads (int, optional): Number of threads to use. Defaults to 25.
        throttle (float or callable, optional): Per-domain throttle in seconds.
            Or a function taking domain name and item and returning the
            throttle to apply. Defaults to 0.2.
        max_redirects (int, optional): Max number of redirections to follow.
        domain_parallelism (int, optional): Max number of urls per domain to
            hit at the same time. Defaults to 1.
        buffer_size (int, optional): Max number of items per domain to enqueue
            into memory in hope of finding a new domain that can be processed
            immediately. Defaults to 1.
        insecure (bool, optional): Whether to ignore SSL certification errors
            when performing requests. Defaults to False.
        timeout (float or urllib3.Timeout, optional): Custom timeout for every
            request.

    Yields:
        FetchWorkerResult

    """

    def generator():
        with FetchThreadPoolExecutor(
            max_workers=threads,
            insecure=insecure,
            timeout=timeout,
            wait=wait,
            daemonic=daemonic
        ) as executor:
            yield from executor.imap_unordered(
                iterator,
                key=key,
                throttle=throttle,
                request_args=request_args,
                buffer_size=buffer_size,
                domain_parallelism=domain_parallelism,
                max_redirects=max_redirects,
                callback=callback
            )

    return generator()


def multithreaded_resolve(iterator, key=None, resolve_args=None, threads=25,
                          throttle=DEFAULT_THROTTLE, max_redirects=5,
                          follow_refresh_header=True, follow_meta_refresh=False,
                          follow_js_relocation=False, infer_redirection=False,
                          buffer_size=DEFAULT_IMAP_BUFFER_SIZE,
                          insecure=False, timeout=DEFAULT_URLLIB3_TIMEOUT,
                          domain_parallelism=DEFAULT_DOMAIN_PARALLELISM,
                          wait=True, daemonic=False):
    """
    Function returning a multithreaded iterator over resolved urls.

    Args:
        iterator (iterable): An iterator over urls or arbitrary items.
        key (callable, optional): Function extracting url from yielded items.
        resolve_args (callable, optional): Function returning specific
            arguments to pass to the resolve util per yielded item.
        threads (int, optional): Number of threads to use. Defaults to 25.
        throttle (float or callable, optional): Per-domain throttle in seconds.
            Or a function taking domain name and item and returning the
            throttle to apply. Defaults to 0.2.
        max_redirects (int, optional): Max number of redirections to follow.
        follow_refresh_header (bool, optional): Whether to follow refresh
            headers. Defaults to True.
        follow_meta_refresh (bool, optional): Whether to follow meta refresh.
            Defaults to False.
        infer_redirection (bool, optional): Whether to infer redirections
            heuristically from urls. Defaults to False.
        buffer_size (int, optional): Max number of items per domain to enqueue
            into memory in hope of finding a new domain that can be processed
            immediately. Defaults to 1.
        insecure (bool, optional): Whether to ignore SSL certification errors
            when performing requests. Defaults to False.
        timeout (float or urllib3.Timeout, optional): Custom timeout for every
            request.
        domain_parallelism (int, optional): Max number of urls per domain to
            hit at the same time. Defaults to 1.

    Yields:
        ResolveWorkerResult

    """

    def generator():
        with ResolveThreadPoolExecutor(
            max_workers=threads,
            insecure=insecure,
            timeout=timeout,
            wait=wait,
            daemonic=daemonic
        ) as executor:
            yield from executor.imap_unordered(
                iterator,
                key=key,
                throttle=throttle,
                resolve_args=resolve_args,
                buffer_size=buffer_size,
                domain_parallelism=domain_parallelism,
                max_redirects=max_redirects,
                follow_refresh_header=follow_refresh_header,
                follow_meta_refresh=follow_meta_refresh,
                follow_js_relocation=follow_js_relocation,
                infer_redirection=infer_redirection
            )

    return generator()



# =============================================================================
# Minet CrowdTangle CLI Utils
# =============================================================================
#
# Miscellaneous generic functions used throughout the CrowdTangle actions.
#
import casanova
import ndjson

from minet.utils import prettyprint_seconds
from minet.cli.utils import print_err, die, LoadingBar
from minet.crowdtangle import CrowdTangleAPIClient
from minet.crowdtangle.exceptions import (
    CrowdTangleInvalidTokenError,
    CrowdTangleRateLimitExceeded,
    CrowdTangleInvalidJSONError
)


def make_paginated_action(method_name, item_name, csv_headers, get_args=None,
                          announce=None):

    def action(cli_args):

        resume = getattr(cli_args, 'resume', False)

        # Validation
        if resume:
            if cli_args.sort_by != 'date':
                die('Cannot --resume if --sort_by is not `date`.')

            if cli_args.format != 'csv':
                die('Cannot --resume jsonl format yet.')

        if cli_args.format == 'csv':
            fieldnames = csv_headers(cli_args) if callable(csv_headers) else csv_headers
            writer = casanova.writer(cli_args.output, fieldnames)
        else:
            writer = ndjson.writer(cli_args.output)

        # Acquiring state from resumer
        if getattr(cli_args, 'resume', False):
            last_date = cli_args.output.pop_state()

            if last_date is not None:
                cli_args.end_date = last_date.replace(' ', 'T')
                print_err('Resuming from: %s' % cli_args.end_date)

        if callable(announce):
            print_err(announce(cli_args))

        # Loading bar
        loading_bar = LoadingBar(
            desc='Fetching %s' % item_name,
            unit=item_name[:-1],
            total=cli_args.limit
        )

        client = CrowdTangleAPIClient(cli_args.token, rate_limit=cli_args.rate_limit)

        args = []

        if callable(get_args):
            args = get_args(cli_args)

        def before_sleep(retry_state):
            exc = retry_state.outcome.exception()

            if isinstance(exc, CrowdTangleRateLimitExceeded):
                reason = 'Call failed because of rate limit!'

            elif isinstance(exc, CrowdTangleInvalidJSONError):
                reason = 'Call failed because of invalid JSON payload!'

            else:
                reason = 'Call failed because of server timeout!'

            loading_bar.print(
                '%s\nWill wait for %s before attempting again.' % (
                    reason,
                    prettyprint_seconds(retry_state.idle_for, granularity=2)
                )
            )

        create_iterator = getattr(client, method_name)
        iterator = create_iterator(
            *args,
            limit=cli_args.limit,
            raw=cli_args.format != 'csv',
            per_call=True,
            detailed=True,
            namespace=cli_args,
            before_sleep=before_sleep
        )

        try:
            for details, items in iterator:
                loading_bar.update(len(items))

                if details is not None:
                    loading_bar.update_stats(**details)

                for item in items:
                    if cli_args.format == 'csv':
                        item = item.as_csv_row()

                    writer.writerow(item)

        except CrowdTangleInvalidTokenError:
            loading_bar.die([
                'Your API token is invalid.',
                'Check that you indicated a valid one using the `--token` argument.'
            ])

    return action



# =============================================================================
# Minet Facebook Formatters
# =============================================================================
#
# Various formatters for Facebook data.
#
from casanova import namedrecord

from minet.facebook.constants import (
    FACEBOOK_COMMENT_CSV_HEADERS,
    FACEBOOK_POST_CSV_HEADERS
)


FacebookComment = namedrecord('FacebookComment', FACEBOOK_COMMENT_CSV_HEADERS)
FacebookPost = namedrecord('FacebookPost', FACEBOOK_POST_CSV_HEADERS)



# =============================================================================
# Minet Facebook Comments Scraping Functions
# =============================================================================
#
# Functions able to scrape comments from Facebook posts.
#
import re
import sys
import soupsieve
from bs4 import BeautifulSoup
from collections import deque
from urllib.parse import urljoin
from ural import force_protocol
from ural.facebook import (
    parse_facebook_url,
    convert_facebook_url_to_mobile,
    has_facebook_comments,
    FacebookGroup as ParsedFacebookGroup
)

from minet.utils import (
    rate_limited_method,
    RateLimiterState,
    parse_date
)
from minet.web import create_pool, request, create_request_retryer
from minet.scrape.std import get_display_text
from minet.facebook.utils import grab_facebook_cookie
from minet.facebook.formatters import FacebookComment, FacebookPost
from minet.facebook.exceptions import (
    FacebookInvalidCookieError,
    FacebookInvalidTargetError
)
from minet.facebook.constants import (
    FACEBOOK_MOBILE_DEFAULT_THROTTLE,
    FACEBOOK_MOBILE_URL
)

VALID_ID_RE = re.compile(r'^\d+$')


def convert_url_to_mobile(url):
    url = force_protocol(url, 'https')
    return convert_facebook_url_to_mobile(url)


def cleanup_post_link(url):
    url = url.replace('//m.', '//www.')

    return url.split('?', 1)[0]


def extract_user_information_from_link(element):
    user_label = element.get_text().strip()
    user_href = element.get('href')
    user = parse_facebook_url(resolve_relative_url(user_href))

    return user_label, user


def resolve_relative_url(url):
    return urljoin(FACEBOOK_MOBILE_URL, url)


def scrape_comments(html, direction=None, in_reply_to=None):
    soup = BeautifulSoup(html, 'lxml')

    data = {
        'direction': direction,
        'post_id': None,
        'comments': [],
        'next': None,
        'replies': [],
        'in_reply_to': in_reply_to
    }

    # Detecting if we are in a video pagelet
    video_pagelet = soup.select_one('#mobile_injected_video_feed_pagelet')

    if video_pagelet is not None:
        actual_comments_link = video_pagelet.select_one('a[href^="/story.php?"]')

        if actual_comments_link:
            data['next'] = resolve_relative_url(actual_comments_link.get('href'))

        return data

    if not in_reply_to:
        if direction is None or direction == 'forward':
            next_link = soup.select_one('[id^="see_next_"] > a[href]')

            if next_link:
                data['next'] = resolve_relative_url(next_link.get('href'))

                if direction is None:
                    data['direction'] = 'forward'

        if direction is None or direction == 'backward':
            next_link = soup.select_one('[id^="see_prev_"] > a[href]')

            if next_link:
                data['next'] = resolve_relative_url(next_link.get('href'))

                if direction is None:
                    data['direction'] = 'backward'
    else:
        if direction is None or direction == 'backward':
            next_link = soup.select_one('[id^="comment_replies_more_1"] > a[href]')

            if next_link:
                data['next'] = resolve_relative_url(next_link.get('href'))

                if direction is None:
                    data['direction'] = 'backward'

    valid_items = (
        item
        for item
        in soup.select('[id]:has(h3 > a)')
        if VALID_ID_RE.match(item.get('id'))
        and not item.parent.get('id', '').startswith('comment_replies_more')
    )

    for item in valid_items:
        item_id = item.get('id')

        # Skipping comment if same as commented
        if item_id == in_reply_to:
            continue

        user_label, user = extract_user_information_from_link(item.select_one('h3 > a'))

        # TODO: link to comment
        content_elements_candidates = item.select_one('h3').find_next_siblings('div')
        content_elements = []
        content_elements_html = []

        for el in content_elements_candidates:
            if el.select_one('[id^=like_]'):
                break

            content_elements_html.append(el)

            if el.get_text().strip():
                content_elements.append(el)

        comment_text = get_display_text(content_elements)
        comment_html = ''.join(str(el) for el in content_elements_html)

        formatted_date = item.select_one('abbr').get_text().strip()
        parsed_date = parse_date(formatted_date)

        post_id_item = item.select_one('[id^="like_"]')

        if post_id_item is None:
            raise TypeError

        post_id = item.select_one('[id^="like_"]').get('id').split('_')[1]

        # NOTE: this could be better (we already know this beforehand)
        data['post_id'] = post_id

        reactions_item = item.select_one('[href^="/ufi/reaction/"]')
        reactions = '0'

        if reactions_item is not None:
            reactions = reactions_item.get_text().strip()

        replies_items = item.select('a[href^="/comment/replies"]')
        replies = '0'

        if len(replies_items) > 0:
            replies_item = replies_items[-1]

            if replies_item is not None:
                replies_text = replies_item.get_text()

                if replies_text != 'Reply':

                    if 'See all' in replies_text:
                        replies = replies_text.split('See all')[-1].split(' replies')[0].strip()
                    else:
                        replies = replies_text.split('·')[-1].split(' repl')[0].strip()

                    replies_url = replies_item.get('href')
                    data['replies'].append((resolve_relative_url(replies_url), item_id))

        data['comments'].append(FacebookComment(
            post_id=post_id,
            id=item_id,
            user_id=getattr(user, 'id', ''),
            user_handle=getattr(user, 'handle', ''),
            user_url=getattr(user, 'url', ''),
            user_label=user_label,
            text=comment_text,
            html=comment_html,
            formatted_date=formatted_date,
            date=parsed_date,
            reactions=reactions,
            replies=replies,
            in_reply_to=in_reply_to
        ))

    return data


def scrape_posts(html):
    soup = BeautifulSoup(html, 'lxml')

    next_link = soup.select_one('a[href*="?bacr="], a[href*="&bacr="]')

    if next_link is not None:
        next_link = resolve_relative_url(next_link.get('href'))

    posts = []
    post_elements = soup.select('#m_group_stories_container > div > [data-ft]')

    for el in post_elements:
        full_story_link = soupsieve.select_one('a:-soup-contains("Full Story")', el)

        if not full_story_link:
            continue

        post_url = cleanup_post_link(full_story_link.get('href'))

        user_label, user = extract_user_information_from_link(el.select_one('h3 a'))

        formatted_date = el.select_one('abbr').get_text().strip()
        parsed_date = parse_date(formatted_date)

        reactions_item = el.select_one('[id^="like_"]')
        reactions = '0'

        if reactions_item:
            reactions_text = reactions_item.get_text()

            if reactions_text.count('·') > 1:
                reactions = reactions_text.split('·', 1)[0].strip()

        comments_item = soupsieve.select_one('a:-soup-contains(" Comment")', el)
        comments = '0'

        if comments_item:
            comments = comments_item.get_text().split('Comment', 1)[0].strip()

        text_root = el.select_one('[data-ft=\'{"tn":"*s"}\']')
        additional_html_roots = []

        img_root = el.select_one('[data-ft=\'{"tn":"H"}\']')

        if img_root:
            additional_html_roots.append(img_root)

        all_text_elements = text_root.find_all('div', recursive=False)

        text_elements = []
        translated_text_elements = []
        translation_link = None

        for text_el in all_text_elements:
            translation_link = text_el.select_one('a[href^="/basic/translation_preferences/"]')
            if translation_link is None:
                text_elements.append(text_el)
            else:
                translation_link.extract()
                translated_text_elements.append(text_el)

        html_elements = text_elements + additional_html_roots

        comment_text = get_display_text(text_elements)
        comment_html = ''.join(str(el) for el in html_elements)

        translated_comment_text = get_display_text(translated_text_elements)
        translated_comment_html = ''.join(str(el) for el in translated_text_elements)
        translated_from = translation_link.get_text().rsplit('from ', 1)[-1].strip() if translation_link else None

        post = FacebookPost(
            url=post_url,
            user_id=getattr(user, 'id', ''),
            user_handle=getattr(user, 'handle', ''),
            user_url=getattr(user, 'url', ''),
            user_label=user_label,
            text=comment_text,
            html=comment_html,
            translated_text=translated_comment_text,
            translated_html=translated_comment_html,
            translated_from=translated_from,
            formatted_date=formatted_date,
            date=parsed_date,
            reactions=reactions,
            comments=comments
        )

        posts.append(post)

    return next_link, posts


class FacebookMobileScraper(object):
    def __init__(self, cookie, throttle=FACEBOOK_MOBILE_DEFAULT_THROTTLE):

        # Grabbing cookie
        cookie = grab_facebook_cookie(cookie)

        if cookie is None:
            raise FacebookInvalidCookieError

        self.cookie = cookie
        self.pool = create_pool()

        self.rate_limiter_state = RateLimiterState(1, throttle)

    @rate_limited_method()
    def request_page(self, url):
        error, result = request(
            url,
            pool=self.pool,
            cookie=self.cookie,
            headers={
                'User-Agent': 'curl/7.68.0',
                'Accept-Language': 'en'
            }
        )

        if error is not None:
            raise error

        return result.data.decode('utf-8')

    def comments(self, url, detailed=False, per_call=False):

        if not has_facebook_comments(url):
            raise FacebookInvalidTargetError

        # Reformatting url to hit mobile website
        url = convert_url_to_mobile(url)

        def generator():
            url_queue = deque([(url, None, None)])

            calls = 0
            replies = 0

            retryer = create_request_retryer()

            while len(url_queue) != 0:
                current_url, direction, in_reply_to = url_queue.popleft()

                html = retryer(self.request_page, current_url)

                try:
                    data = scrape_comments(html, direction, in_reply_to)
                except TypeError:
                    # with open('./dump.html', 'w') as f:
                    #     f.write(html)
                    print('Could not process comment in %s' % current_url, file=sys.stderr)
                    return

                calls += 1

                for reply_url, commented_id in data['replies']:
                    url_queue.append((reply_url, None, commented_id))

                if data['next'] is not None:
                    url_queue.append((data['next'], data['direction'], in_reply_to))

                comments = data['comments']

                for comment in data['comments']:
                    if in_reply_to is not None:
                        replies += 1

                    if not per_call:
                        yield comment

                if per_call and len(comments) > 0:
                    if detailed:
                        details = {
                            'calls': calls,
                            'replies': replies,
                            'queue_size': len(url_queue)
                        }

                        yield details, comments
                    else:
                        yield comments

        return generator()

    def posts(self, url):
        parsed = parse_facebook_url(url)

        if not isinstance(parsed, ParsedFacebookGroup):
            raise FacebookInvalidTargetError

        url = convert_url_to_mobile(parsed.url)

        def generator():

            retryer = create_request_retryer()
            current_url = url

            while True:
                html = retryer(self.request_page, current_url)

                # with open('./dump.html', 'w') as f:
                #     f.write(html)

                next_url, posts = scrape_posts(html)

                for post in posts:
                    yield post

                if next_url is None or len(posts) == 0:
                    break

                current_url = next_url

        return generator()



# =============================================================================
# Minet Custom Exceptions
# =============================================================================
#
# Collection of handy custom exceptions.
#


# Base minet error
class MinetError(Exception):
    def __init__(self, message=None):
        super().__init__('' if message is None else message)
        self.message = message

    def __repr__(self):
        representation = '<' + self.__class__.__name__

        for k in dir(self):
            if (
                k.startswith('_') or
                k == 'args' or
                'traceback' in k
            ):
                continue

            v = getattr(self, k)

            if v is None:
                continue

            if isinstance(v, BaseException):
                representation += ' {k}=<{name}>'.format(k=k, name=v.__class__.__name__)
                continue

            if isinstance(v, str) and len(v) > 30:
                v = v[:29] + '…'

            representation += ' {k}={v!r}'.format(k=k, v=v)

        representation += '>'

        return representation


# General errors
class UnknownEncodingError(MinetError):
    pass


# Miscellaneous HTTP errors
class InvalidURLError(MinetError):
    def __init__(self, message=None, url=None):
        self.url = url
        super().__init__(message)


# Redirection errors
class RedirectError(MinetError):
    pass


class MaxRedirectsError(RedirectError):
    pass


class InfiniteRedirectsError(RedirectError):
    pass


class SelfRedirectError(InfiniteRedirectsError):
    pass


class InvalidRedirectError(RedirectError):
    pass


# Definition errors
class DefinitionInvalidFormatError(MinetError):
    pass


# Crawling errors
class CrawlError(MinetError):
    pass


class UnknownSpiderError(CrawlError):
    pass


# Extraction errors
class TrafilaturaError(MinetError):
    def __init__(self, msg=None, reason=None):
        super().__init__(msg)
        self.reason = reason


# Filesystem errors
class FilenameFormattingError(MinetError):
    def __init__(self, msg=None, reason=None, template=None):
        super().__init__(msg)
        self.reason = reason
        self.template = template



# =============================================================================
# Minet Scrape Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class ScrapeError(MinetError):
    pass


class InvalidScraperError(ScrapeError):
    def __init__(self, msg=None, validation_errors=[]):
        super().__init__(msg)
        self.validation_errors = validation_errors


class CSSSelectorTooComplex(ScrapeError):
    pass


class ScraperNotTabularError(ScrapeError):
    pass


class ScraperRuntimeError(ScrapeError):
    def __init__(self, msg=None, reason=None, expression=None, path=None):
        super().__init__(msg)
        self.reason = reason
        self.expression = expression
        self.path = path


class ScraperEvalError(ScraperRuntimeError):
    pass


class NotATableError(ScraperRuntimeError):
    pass


class ScraperEvalSyntaxError(ScraperRuntimeError):
    pass


class ScraperEvalTypeError(ScraperRuntimeError):
    def __init__(self, msg=None, expected=None, got=None, **kwargs):
        super().__init__(msg, **kwargs)
        self.expected = expected
        self.got = got


class ScraperEvalNoneError(ScraperRuntimeError):
    pass


class ScraperValidationError(ScraperRuntimeError):
    pass


class ScraperValidationMixedConcernError(ScraperRuntimeError):
    pass


class ScraperValidationConflictError(ScraperRuntimeError):
    def __init__(self, msg=None, keys=[], **kwargs):
        super().__init__(msg, **kwargs)
        self.keys = keys


class ScraperValidationIrrelevantPluralModifierError(ScraperRuntimeError):
    def __init__(self, msg=None, modifier=None, **kwargs):
        super().__init__(msg, **kwargs)
        self.modifier = modifier


class ScraperValidationInvalidPluralModifierError(ScraperRuntimeError):
    def __init__(self, msg=None, modifier=None, **kwargs):
        super().__init__(msg, **kwargs)
        self.modifier = modifier


class ScraperValidationInvalidExtractorError(ScraperRuntimeError):
    def __init__(self, msg=None, extractor=None, **kwargs):
        super().__init__(msg, **kwargs)
        self.extractor = extractor


class InvalidCSSSelectorError(ScraperRuntimeError):
    pass
